import json
import logging
import pathlib
import aiohttpx
import contextlib
from typing import Optional, Dict, Union, Any, List, Type, TYPE_CHECKING
from lazyops.types import BaseSettings, validator, BaseModel, lazyproperty, Field, PYD_VERSION
from lazyops.libs.proxyobj import ProxyObject
from lazyops.libs.abcs.configs.types import AppEnv
from async_openai.version import VERSION
from async_openai.types.options import ApiType
from async_openai.types.context import ModelCostItem, ModelContextHandler

if PYD_VERSION == 2:
    from pydantic import model_validator
else:
    from lazyops.types.models import root_validator

if TYPE_CHECKING:
    from async_openai.types.routes import BaseRoute, BaseResource


preset_path = pathlib.Path(__file__).parent.joinpath('presets')


class ExternalProviderConfig(BaseModel):
    """
    External Provider - Configuration
    """
    api_base: str = Field(..., description="The base URL for the API")
    api_path: Optional[str] = Field(None, description="The path for the API")
    api_key_header: Optional[str] = Field(None, description="The header for the API Key")
    api_key_scheme: Optional[str] = Field(None, description="The scheme for the API Key")
    api_key: Optional[str] = Field(None, description="The API Key")

    custom_headers: Optional[Dict[str, str]] = Field(None, description="Custom Headers")
    proxy_url: Optional[str] = Field(None, description="The Proxy URL")
    proxy_headers: Optional[Dict[str, str]] = Field(None, description="Proxy Headers")
    hf_compatible: Optional[bool] = Field(None, description="Whether the provider is HuggingFace Compatible for Tokenization")

    @validator("custom_headers", "proxy_headers", pre=True)
    def validate_headers(cls, value: Optional[Dict[str, Any]]) -> Optional[Dict[str, str]]:
        """
        Validates the Headers
        """
        if value is None: return {}
        return {str(k).strip(): str(v).strip() for k,v in value.items()}

    @property
    def has_api_key(self) -> bool:
        """
        Returns whether the API Key is present
        """
        return self.api_key is not None
    
    @property
    def api_key_value(self) -> Optional[str]:
        """
        Returns the API Key Value
        """
        if not self.has_api_key: return None
        return f"{self.api_key_scheme} {self.api_key}" if self.api_key_scheme else self.api_key
    
    @property
    def has_proxy(self) -> bool:
        """
        Returns whether the Proxy is present
        """
        return self.proxy_url is not None
    
    @property
    def api_url(self) -> str:
        """
        Returns the API URL
        """
        if not self.api_path: return self.api_base
        return aiohttpx.urljoin(self.api_base, self.api_path)


class ProviderModel(ModelCostItem):
    """
    A Provider Model
    """
    name: str = Field(..., description="The Model Name")


class ProviderRoute(BaseModel):
    """
    A Provider Route
    """
    name: Optional[str] = Field(None, description="The Route Name")
    api_resource: Optional[str] = Field(None, description="The Route Path")
    root_name: Optional[str] = Field(None, description="The Root Name")
    # params: Optional[Dict[str, Type]] = Field(None, description="The Route Parameters and expected Types")

    if TYPE_CHECKING:
        route_class: Optional[Type['BaseRoute']] = Field(None, description="The Route Class")
        object_class: Optional[Type['BaseResource']] = Field(None, description="The Resource Class")
        response_class: Optional[Type['BaseResource']] = Field(None, description="The Response Class")
    else:
        route_class: Optional[Any] = Field(None, description="The Route Class")
        object_class: Optional[Any] = Field(None, description="The Resource Class")
        response_class: Optional[Any] = Field(None, description="The Response Class")
    

    @validator("route_class", "object_class", "response_class", pre=True)
    def validate_route_classes(cls, value: Union[str, Type]) -> Type:
        """
        Validates the Route Classes
        """
        if value is None: return None
        if isinstance(value, str):
            from lazyops.utils.lazy import lazy_import
            return lazy_import(value)
        return value


class ExternalProviderRoutes(BaseModel):
    """
    External Provider - Routes
    """
    completion: Optional[ProviderRoute] = Field(None, description="The Completion Route")
    chat: Optional[ProviderRoute] = Field(None, description="The Chat Route")
    embedding: Optional[ProviderRoute] = Field(None, description="The Embedding Route")
    # if TYPE_CHECKING:

    if PYD_VERSION == 2:
        @model_validator(mode = 'after')
        def validate_routes(self):
            """
            Validates the Routes
            """
            from async_openai.schemas import (
                ChatRoute, ChatResponse, ChatObject,
                CompletionRoute, CompletionResponse, CompletionObject,
                EmbeddingRoute, EmbeddingResponse, EmbeddingObject,
                # TODO - Add the rest of the routes
            )
            if self.chat is None:
                self.chat = ProviderRoute(
                    name = "chat",
                    route_class = ChatRoute,
                    object_class = ChatObject,
                    response_class = ChatResponse,
                )
            if not self.chat.name:
                self.chat.name = "chat"
            if self.chat.route_class is None:
                self.chat.route_class = ChatRoute
            if self.chat.object_class is None:
                self.chat.object_class = ChatObject
            if self.chat.response_class is None:
                self.chat.response_class = ChatResponse

            if self.completion is None:
                self.completion = ProviderRoute(
                    name = "completion",
                    route_class = CompletionRoute,
                    object_class = CompletionObject,
                    response_class = CompletionResponse,
                )
            if not self.completion.name:
                self.completion.name = "completion"
            if self.completion.route_class is None:
                self.completion.route_class = CompletionRoute
            if self.completion.object_class is None:
                self.completion.object_class = CompletionObject
            if self.completion.response_class is None:
                self.completion.response_class = CompletionResponse

            if self.embedding is None:
                self.embedding = ProviderRoute(
                    name = "embedding",
                    route_class = EmbeddingRoute,
                    object_class = EmbeddingObject,
                    response_class = EmbeddingResponse,
                )
            
            if not self.embedding.name:
                self.embedding.name = "embedding"
            if self.embedding.route_class is None:
                self.embedding.route_class = EmbeddingRoute
            if self.embedding.object_class is None:
                self.embedding.object_class = EmbeddingObject
            if self.embedding.response_class is None:
                self.embedding.response_class = EmbeddingResponse
            return self

    else:
        @root_validator()
        def validate_routes(cls, values: Dict[str, Any]) -> Dict[str, Any]:
            """
            Validates the Routes
            """
            from async_openai.schemas import (
                ChatRoute, ChatResponse, ChatObject,
                CompletionRoute, CompletionResponse, CompletionObject,
                EmbeddingRoute, EmbeddingResponse, EmbeddingObject,
                # TODO - Add the rest of the routes
            )
            if values.get('chat') is None:
                values['chat'] = ProviderRoute(
                    name = "chat",
                    route_class = ChatRoute,
                    object_class = ChatObject,
                    response_class = ChatResponse,
                )
            else:
                values['chat'] = ProviderRoute.parse_obj(values['chat'])
            if not values['chat'].name:
                values['chat'].name = "chat"
            if values['chat'].route_class is None:
                values['chat'].route_class = ChatRoute
            if values['chat'].object_class is None:
                values['chat'].object_class = ChatObject
            if values['chat'].response_class is None:
                values['chat'].response_class = ChatResponse
            
            if values.get('completion') is None:
                values['completion'] = ProviderRoute(
                    name = "completion",
                    route_class = CompletionRoute,
                    object_class = CompletionObject,
                    response_class = CompletionResponse,
                )
            else:
                values['completion'] = ProviderRoute.parse_obj(values['completion'])
            if not values['completion'].name:
                values['completion'].name = "completion"
            if values['completion'].route_class is None:
                values['completion'].route_class = CompletionRoute
            if values['completion'].object_class is None:
                values['completion'].object_class = CompletionObject
            if values['completion'].response_class is None:
                values['completion'].response_class = CompletionResponse
            
            if values.get('embedding') is None:
                values['embedding'] = ProviderRoute(
                    name = "embedding",
                    route_class = EmbeddingRoute,
                    object_class = EmbeddingObject,
                    response_class = EmbeddingResponse,
                )
            else:
                values['embedding'] = ProviderRoute.parse_obj(values['embedding'])
            if not values['embedding'].name:
                values['embedding'].name = "embedding"
            if values['embedding'].route_class is None:
                values['embedding'].route_class = EmbeddingRoute
            if values['embedding'].object_class is None:
                values['embedding'].object_class = EmbeddingObject
            if values['embedding'].response_class is None:
                values['embedding'].response_class = EmbeddingResponse
            return values
        
    @property
    def api_route_classes(self) -> Dict[str, Type['BaseRoute']]:
        """
        Returns the Route Classes
        """
        return {
            "chat": self.chat.route_class,
            "completion": self.completion.route_class,
            "embedding": self.embedding.route_class,
        }
            

class ExternalProviderSettings(BaseModel):
    """
    External Provider Configuration
    """
    name: str = Field(..., description="The Provider Name")
    config: ExternalProviderConfig = Field(default_factory = ExternalProviderConfig, description="The Provider Configuration")
    models: List[ProviderModel] = Field(default_factory=list, description="The Provider Models")
    routes: Optional[ExternalProviderRoutes] = Field(default_factory=ExternalProviderRoutes, description="The Provider Routes")

    @classmethod
    def from_preset(
        cls, 
        name: Optional[str] = None,
        path: Optional[Union[str, pathlib.Path]] = None,
        **overrides
    ) -> 'ExternalProviderSettings':
        """
        Loads the Provider Settings from a Preset
        """
        assert name or path, "You must provide either a name or a path to the preset"
        if path:
            preset_file = pathlib.Path(path)
            assert preset_file.exists(), f"Could not find the preset path: {preset_file}"
        else:
            preset_file = preset_path.joinpath(f"{name}.yaml")
            if not preset_file.exists():
                raise FileNotFoundError(f"Could not find the preset file: {preset_file} for {name}")
        
        assert preset_file.suffix in {
            ".yaml", ".yml", ".json"
        }, f"The preset file must be a YAML or JSON file: {preset_file}"
        
        from lazyops.libs.abcs.utils.envvars import parse_envvars_from_text
        
        text = preset_file.read_text()
        text, _ = parse_envvars_from_text(text)
        if preset_file.suffix == ".json":
            data = json.loads(text)
        else:
            import yaml
            data = yaml.safe_load(text)
        
        if overrides: data.update(overrides)
        provider_settings = cls.parse_obj(data)
        ModelContextHandler.add_provider(provider_settings)
        return provider_settings
    
    @property
    def model_list(self) -> List[str]:
        """
        Returns the Model List (excludes the provider name)
        """
        names = [m.name for m in self.models]
        # Add aliases
        for m in self.models:
            if m.aliases:
                names.extend(m.aliases)
        return names


class ExternalProviderAuth(aiohttpx.Auth):
    """
    Custom Authentication Wrapper for External OpenAI Clients
    """
    def __init__(
        self, 
        config: ExternalProviderConfig,
        is_proxied: Optional[bool] = None,
        **kwargs,
    ):
        """
        Initializes the External Provider Auth

        :TODO - add support for Proxy
        """
        self.config = config
        self.is_proxied = is_proxied

    def auth_flow(self, request):
        """
        Injects the API Key into the Request
        """
        if self.config.has_api_key and self.config.api_key_header not in request.headers:
            request.headers[self.config.api_key_header] = self.config.api_key_value
        if self.config.custom_headers:
            request.headers.update(self.config.custom_headers)        
        yield request

    async def async_auth_flow(self, request):
        """
        Injects the API Key into the Request
        """
        if self.config.has_api_key and self.config.api_key_header not in request.headers:
            request.headers[self.config.api_key_header] = self.config.api_key_value
        if self.config.custom_headers:
            request.headers.update(self.config.custom_headers)
        yield request
        
        


from __future__ import annotations

"""
Together.xyz Chat Route
"""
import json
from ...chat import (
    ChatRoute as BaseChatRoute,
    ChatObject as BaseChatObject,
    ChatResponse as BaseChatResponse,
    ChatChoice as BaseChatChoice,
    ChatMessage as BaseChatMessage, 
    Function, FunctionCall, Tool, logger
)
from lazyops.types import validator, root_validator, BaseModel, lazyproperty, Field, PYD_VERSION
from async_openai.types.context import ModelContextHandler
from typing import Any, Dict, List, Optional, Union, Set, Type, TYPE_CHECKING

# if PYD_VERSION == 2:
#     from pydantic import model_validator
# else:
#     from lazyops.types.models import root_validator

class ChatObject(BaseChatObject):
    model: Optional[str] = "mistralai/Mistral-7B-Instruct-v0.1"
    response_format: Optional[Dict[str, Union[str, Dict[str, Any]]]] = None
    is_json_mode: Optional[bool] = Field(None, exclude = True)

    @validator('model', pre=True, always=True)
    def validate_model(cls, v, values: Dict[str, Any]) -> str:
        """
        Validate the model
        """
        if not v:
            if values.get('engine'):
                v = values.get('engine')
            elif values.get('deployment'):
                v = values.get('deployment')
        
        v = ModelContextHandler.resolve_external_model_name(v)
        return v


    """
    Handle Validation for JSON Mode

    JSON mode corrals the LLM into outputting JSON conforming to a provided schema. 
    To activate JSON mode, provide the response_format parameter to the Chat Completions 
    API with {"type": "json_object"}. The JSON Schema can be specified with the schema 
    property of response_format. The schema property should be a JSON Schema object.
    """

    # if PYD_VERSION == 2:
    #     @model_validator(mode = 'after')
    #     def validate_obj(self):
    #         """
    #         Validate the object
    #         """
        
    # else:

    @root_validator(pre = True)
    def validate_obj(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """
        Validate the object
        """
        if values.get('functions'):
            if not all(isinstance(f, Function) for f in values['functions']):
                values['functions'] = [Function(**f) for f in values['functions']]
            if not values.get('function_call'):
                values['function_call'] = 'auto'
        
        if values.get('response_format', {}).get('type') == 'json_object':
            values['is_json_mode'] = True
            if not values['response_format'].get('schema') and  values.get('functions'):
                func = values['functions'][0] if \
                    len(values['functions']) == 1 or \
                    values.get('function_call') == 'auto' else \
                    next((f for f in values['functions'] if f.name == values['function_call']))
                
                assert func, 'No function found'
                schema = func.model_json_schema()
                # _ = schema.pop('additionalProperties', None)
                values['response_format']['schema'] = schema
                # logger.info(values["response_format"], prefix = 'JSON Mode')

                
        # Disable tools if response format is json_object
        elif values.get('tools'):
            tools = []
            for tool in values['tools']:
                if isinstance(tool, Tool):
                    tools.append(tool)
                elif isinstance(tool, dict):
                    # This should be the correct format
                    if tool.get('function'):
                        tools.append(Tool(**tool))
                    else:
                        # This is previously supported format
                        tools.append(Tool(function = Function(**tool)))
                else:
                    raise ValueError(f'Invalid tool: {tool}')
            values['tools'] = tools
            if not values.get('tool_choice'):
                values['tool_choice'] = 'auto'
        return values
    
    def dict(self, **kwargs) -> Dict[str, Any]:
        """
        Return the dict
        """
        exclude: Set[str] = kwargs.pop('exclude', None) or set()
        if self.is_json_mode:
            exclude.add('tools')
            exclude.add('tool_choice')
            exclude.add('functions')
            exclude.add('function_call')
        
        return super().dict(exclude = exclude, **kwargs)
        

class ChatMessage(BaseChatMessage):
    
    """
    Handle some validation here
    """

    @root_validator(pre = True)
    def validate_message(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """
        Validate the object
        """
        # if values.get('content') and '"arguments"' in values['content']:
        #     content = values.pop('content')
        #     try:
        #         data = json.loads(content)
        #     except Exception as e:
        #         try:
        #             content = content.split('\n', 1)[-1].strip()
        #             data = json.loads(content)
        #         except Exception as e:
        #             logger.error(f'Invalid JSON: {content}: {e}')
        #             raise e
        #     values['function_call'] = FunctionCall(**data)
        if values.get('tool_calls'):
            for tc in values['tool_calls']:
                if tc.get('type') == 'function' and tc.get('function'):
                    func = FunctionCall(**tc['function'])
                    values['function_call'] = func
                    break
        return values



class ChatChoice(BaseChatChoice):
    message: ChatMessage


class ChatResponse(BaseChatResponse):

    input_object: Optional[ChatObject] = None
    choice_model: Optional[Type[ChatChoice]] = ChatChoice


class ChatRoute(BaseChatRoute):
    input_model: Optional[Type[ChatObject]] = ChatObject
    response_model: Optional[Type[ChatResponse]] = ChatResponse


import enum
import json
import time
import asyncio
import aiohttpx
import contextlib
from typing import Optional, Type, Any, Union, List, Dict, Iterator, AsyncIterator, Generator, AsyncGenerator, TYPE_CHECKING
from lazyops.types import validator, lazyproperty
from lazyops.types.models import root_validator, pre_root_validator, Field

from async_openai.types.context import ModelContextHandler
from async_openai.types.resources import BaseResource, Usage
from async_openai.types.responses import BaseResponse
from async_openai.types.routes import BaseRoute
from async_openai.types.errors import RateLimitError, APIError, MaxRetriesExceeded, InvalidMaxTokens, InvalidRequestError
from async_openai.utils import logger, parse_stream, aparse_stream


__all__ = [
    'CompletionChoice',
    'CompletionObject',
    'CompletionResponse',
    'CompletionRoute',
]


class CompletionKind(str, enum.Enum):
    # Only have one choice
    CONTENT = 'content'

class StreamedCompletionChoice(BaseResource):
    kind: CompletionKind = CompletionKind.CONTENT
    value: str


class CompletionChoice(BaseResource):
    text: str
    index: int
    logprobs: Optional[Any] = None
    finish_reason: Optional[str] = None



class CompletionObject(BaseResource):
    prompt: Union[List[str], str] = '<|endoftext|>'
    model: Optional[str] = "gpt-3.5-turbo-instruct"
    # prompt: Optional[Union[List[str], str]] = '<|endoftext|>'
    suffix: Optional[str] = None
    max_tokens: Optional[int] = 16
    temperature: Optional[float] = 1.0
    top_p: Optional[float] = 1.0
    n: Optional[int] = 1
    stream: Optional[bool] = False
    logprobs: Optional[int] = None
    echo: Optional[bool] = False
    stop: Optional[Union[List[str], str]] = None
    presence_penalty: Optional[float] = 0.0
    frequency_penalty: Optional[float] = 0.0
    best_of: Optional[int] = 1
    logit_bias: Optional[Dict[str, float]] = None
    user: Optional[str] = None

    validate_max_tokens: Optional[bool] = Field(default = False, exclude = True)
    # validate_model_aliases: Optional[bool] = Field(default = False, exclude = True)

    @validator('model', pre=True, always=True)
    def validate_model(cls, v, values: Dict[str, Any]) -> str:
        """
        Validate the model
        """
        if not v:
            if values.get('engine'):
                v = values.get('engine')
            elif values.get('deployment'):
                v = values.get('deployment')
        v = ModelContextHandler.resolve_model_name(v)
        # if values.get('validate_model_aliases', False):
        #     v = ModelContextHandler[v].name
        return v

    
    @validator('temperature')
    def validate_temperature(cls, v: float) -> float:
        """
        Min Temperature is 0.0
        https://beta.openai.com/docs/api-reference/completions/create#completions/create-temperature
        """
        return None if v is None else max(0.0, v)
    
    @validator('top_p')
    def validate_top_p(cls, v: float) -> float:
        """
        Min Top Probability is 0.0
        https://beta.openai.com/docs/api-reference/completions/create#completions/create-top-probability
        """
        return None if v is None else max(0.0, v)

    @validator('logprobs')
    def validate_logprobs(cls, v: int) -> int:
        """
        Max logprobs is 5
        https://beta.openai.com/docs/api-reference/completions/create#completions/create-logprobs
        """
        return None if v is None else max(0, min(v, 5))
    
    @validator('presence_penalty')
    def validate_presence_penalty(cls, v: float) -> float:
        """
        Min Presence Penalty is -2.0, Max is 2.0
        https://beta.openai.com/docs/api-reference/completions/create#completions/create-presence-penalty
        """
        return None if v is None else max(0.0, min(v, 2.0))
    
    @validator('frequency_penalty')
    def validate_frequency_penalty(cls, v: float) -> float:
        """
        Min Frequency Penalty is -2.0, Max is 2.0
        https://beta.openai.com/docs/api-reference/completions/create#completions/create-frequency-penalty
        """
        return None if v is None else max(0.0, min(v, 2.0))
    

    def dict(self, *args, exclude: Any = None, **kwargs):
        """
        Returns the dict representation of the response
        """
        return super().dict(*args, exclude = exclude, **kwargs)


    

class CompletionResponse(BaseResponse):
    choices: Optional[List[CompletionChoice]] = None
    choice_model: Optional[Type[BaseResource]] = CompletionChoice
    input_object: Optional[CompletionObject] = None

    @lazyproperty
    def text(self) -> str:
        """
        Returns the text for the completions
        """
        if self.choices_results:
            return ''.join([choice.text for choice in self.choices])
        return self.response.text
    
    @lazyproperty
    def openai_model(self):
        """
        Returns the model for the completions
        """
        return self.headers.get('openai-model', self.completion_model)

    @lazyproperty
    def completion_model(self):
        """
        Returns the model for the completions
        """
        return self.input_object.model or None

    def _validate_usage(self):
        """
        Validate usage
        """
        if self.usage and self.usage.total_tokens: return
        if self.response.status_code == 200:
            self.usage = Usage(
                prompt_tokens = ModelContextHandler.count_tokens(self.input_object.prompt),
                completion_tokens = ModelContextHandler.count_tokens(self.text),
            )
            self.usage.total_tokens = self.usage.prompt_tokens + self.usage.completion_tokens

    @lazyproperty
    def consumption(self) -> int:
        """
        Returns the consumption for the completions
        """ 
        self._validate_usage()
        return ModelContextHandler.get_consumption_cost(
            model_name = self.openai_model,
            usage = self.usage,
        )

    def dict(self, *args, exclude: Any = None, **kwargs):
        """
        Returns the dict representation of the response
        """
        return super().dict(*args, exclude = exclude, **kwargs)

    def parse_stream_item(self, item: Union[Dict, Any], **kwargs) -> Optional[StreamedCompletionChoice]:
        """
        Parses a single stream item
        """
        choice = item['choices'][0]
        if choice['finish_reason'] == 'stop':
            return None
        return StreamedCompletionChoice(
            value = choice['text']
        )

    def handle_stream(
        self,
        response: aiohttpx.Response,
        streaming: Optional[bool] = False,
    ) -> Iterator[Dict]:
        """
        Handles the streaming response
        """
        texts = {}
        for line in parse_stream(response):
            try:
                item = json.loads(line)
                if streaming:
                    yield item
                self.handle_stream_metadata(item)
                for n, choice in enumerate(item['choices']):
                    if not texts.get(n):
                        texts[n] = {
                            'index': choice['index'],
                            'text': choice['text'],
                        }
                        self.usage.completion_tokens += 1
                    elif choice['finish_reason'] != 'stop':
                        texts[n]['text'] += choice['text']
                        self.usage.completion_tokens += 1

                    else:
                        texts[n]['finish_reason'] = choice['finish_reason']
                        compl = texts.pop(n)
                        if streaming:
                            self.handle_resource_item(item = compl)
                        else:
                            yield compl

            except Exception as e:
                logger.error(f'Error: {line}: {e}')

        self._stream_consumed = True
        for remaining_text in texts.values():
            if streaming:
                self.handle_resource_item(item = remaining_text)
            else:
                yield remaining_text
        self.usage.total_tokens = self.usage.completion_tokens
    

    async def ahandle_stream(
        self,
        response: aiohttpx.Response,
        streaming: Optional[bool] = False,
    ) -> AsyncIterator[Dict]:
        """
        Handles the streaming response
        """
        texts = {}
        async for line in aparse_stream(response):
            try:
                item = json.loads(line)
                if streaming:
                    yield item
                self.handle_stream_metadata(item)
                for n, choice in enumerate(item['choices']):
                    if not texts.get(n):
                        texts[n] = {
                            'index': choice['index'],
                            'text': choice['text'],
                        }
                        self.usage.completion_tokens += 1
                    elif choice['finish_reason'] != 'stop':
                        texts[n]['text'] += choice['text']
                        self.usage.completion_tokens += 1

                    else:
                        texts[n]['finish_reason'] = choice['finish_reason']
                        compl = texts.pop(n)
                        if streaming:
                            self.handle_resource_item(item = compl)
                        else:
                            yield compl

            except Exception as e:
                logger.error(f'Error: {line}: {e}')

        self._stream_consumed = True
        for remaining_text in texts.values():
            if streaming:
                self.handle_resource_item(item = remaining_text)
            else:
                yield remaining_text
        self.usage.total_tokens = self.usage.completion_tokens

    if TYPE_CHECKING:
        def stream(self, **kwargs) -> Generator[StreamedCompletionChoice, None, None]:
            """
            Streams the response
            """
            ...

        async def astream(self, **kwargs) -> AsyncGenerator[StreamedCompletionChoice, None]:
            """
            Streams the response
            """
            ...
    

class CompletionRoute(BaseRoute):
    input_model: Optional[Type[BaseResource]] = CompletionObject
    response_model: Optional[Type[BaseResource]] = CompletionResponse

    api_resource: Optional[str] = Field(default = 'completions')
    root_name: Optional[str] = Field(default = 'completion')

    # @lazyproperty
    # def api_resource(self):
    #     return 'completions'

    # @lazyproperty
    # def root_name(self):
    #     return 'completion'
    
    def create(
        self, 
        input_object: Optional[CompletionObject] = None,
        parse_stream: Optional[bool] = True,
        auto_retry: Optional[bool] = False,
        auto_retry_limit: Optional[int] = None,
        **kwargs
    ) -> CompletionResponse:
        """
        Creates a completion for the provided prompt and parameters

        Usage:

        ```python
        >>> result = OpenAI.completions.create(
        >>>    prompt = 'say this is a test',
        >>>    max_tokens = 4,
        >>>    stream = True
        >>> )
        ```

        **Parameters:**

        :model (required): ID of the model to use. You can use the List models API 
        to see all of your available models,  or see our Model overview for descriptions of them.
        Default: `text-curie-001`
        
        :prompt (optional): The prompt(s) to generate completions for, encoded as 
        a string, array of strings, array of tokens, or array of token arrays.
        Note that `<|endoftext|>` is the document separator that the model sees during 
        training, so if a prompt is not specified the model will generate as if from 
        the beginning of a new document.
        Default: `<|endoftext|>`

        :suffix (optional): The suffix that comes after a completion of inserted text.
        Default: `None`

        :max_tokens (optional): The maximum number of tokens to generate in the completion.
        The token count of your prompt plus `max_tokens` cannot exceed the model's context length. 
        Most models have a context length of 2048 tokens (except for the newest models, which 
        support 4096).
        Default: `16`

        :temperature (optional): What sampling temperature to use. Higher values means 
        the model will take more risks. Try 0.9 for more creative applications, and 0 (argmax sampling) 
        for ones with a well-defined answer. We generally recommend altering this or `top_p` but not both.
        Default: `1.0`

        :top_p (optional): An alternative to sampling with `temperature`, called nucleus 
        sampling, where the model considers the results of the tokens with `top_p` probability mass. 
        So `0.1` means only  the tokens comprising the top 10% probability mass are considered.
        We generally recommend altering this or `temperature` but not both
        Default: `1.0`

        :n (optional): How many completions to generate for each prompt.
        Note: Because this parameter generates many completions, it can quickly 
        consume your token quota. Use carefully and ensure that you have reasonable 
        settings for `max_tokens` and stop.
        Default: `1`

        :stream (optional): Whether to stream back partial progress. 
        If set, tokens will be sent as data-only server-sent events as they become 
        available, with the stream terminated by a `data: [DONE]` message. This is 
        handled automatically by the Client and enables faster response processing.
        Default: `False`

        :logprobs (optional): Include the log probabilities on the `logprobs` 
        most likely tokens, as well the chosen tokens. For example, if `logprobs` is 5, 
        the API will return a list of the 5 most likely tokens. The API will always 
        return the logprob of the sampled token, so there may be up to `logprobs+1` 
        elements in the response. The maximum value for `logprobs` is 5.
        Default: `None`

        :echo (optional): Echo back the prompt in addition to the completion
        Default: `False`

        :stop (optional): Up to 4 sequences where the API will stop generating 
        further tokens. The returned text will not contain the stop sequence.
        Default: `None`

        :presence_penalty (optional): Number between `-2.0` and `2.0`. Positive values 
        penalize new tokens based on whether they appear in the text so far, increasing the 
        model's likelihood to talk about new topics
        Default: `0.0`

        :frequency_penalty (optional): Number between `-2.0` and `2.0`. Positive 
        values penalize new tokens based on their existing frequency in the text so 
        far, decreasing the model's likelihood to repeat the same line verbatim.
        Default: `0.0`

        :best_of (optional): Generates `best_of` completions server-side and returns 
        the "best" (the one with the highest log probability per token). Results cannot be streamed.
        When used with `n`, `best_of` controls the number of candidate completions and n specifies how 
        many to return – `best_of` must be greater than `n`.
        Note: Because this parameter generates many completions, it can quickly consume your token quota. 
        Use carefully and ensure that you have reasonable settings for `max_tokens` and `stop`.
        Default: `1`

        :logit_bias (optional): Modify the likelihood of specified tokens appearing in the completion.
        Accepts a json object that maps tokens (specified by their token ID in the GPT tokenizer) to an associated 
        bias value from -100 to 100. You can use this tokenizer tool (which works for both GPT-2 and GPT-3) to 
        convert text to token IDs. Mathematically, the bias is added to the logits generated by the model prior 
        to sampling. The exact effect will vary per model, but values between -1 and 1 should decrease or increase 
        likelihood of selection; values like -100 or 100 should result in a ban or exclusive selection of the 
        relevant token.
        As an example, you can pass `{"50256": -100}` to prevent the `<|endoftext|>` token from being generated.
        Default: `None`

        :user (optional): A unique identifier representing your end-user, which can help OpenAI to 
        monitor and detect abuse.
        Default: `None`

        :auto_retry (optional): Whether to automatically retry the request if it fails due to a rate limit error.

        :auto_retry_limit (optional): The maximum number of times to retry the request if it fails due to a rate limit error.

        Returns: `CompletionResult`
        """
        current_attempt = kwargs.pop('_current_attempt', 0)
        if not auto_retry:
            return super().create(input_object = input_object, parse_stream = parse_stream, **kwargs)
        
        # Handle Auto Retry Logic
        if not auto_retry_limit: auto_retry_limit = self.settings.max_retries
        try:
            return super().create(input_object = input_object, parse_stream = parse_stream, **kwargs)
        except RateLimitError as e:
            if current_attempt >= auto_retry_limit:
                raise MaxRetriesExceeded(name = self.name, attempts = current_attempt, base_exception = e) from e
            sleep_interval = e.retry_after_seconds * 1.5 if e.retry_after_seconds else 15.0
            logger.warning(f'[{self.name}: {current_attempt}/{auto_retry_limit}] Rate Limit Error. Sleeping for {sleep_interval} seconds')
            time.sleep(sleep_interval)
            current_attempt += 1
            return self.create(
                input_object = input_object,
                parse_stream = parse_stream,
                auto_retry = auto_retry,
                auto_retry_limit = auto_retry_limit,
                _current_attempt = current_attempt,
                **kwargs
            )
        except APIError as e:
            if current_attempt >= auto_retry_limit:
                raise MaxRetriesExceeded(name = self.name, attempts = current_attempt, base_exception = e) from e
            logger.warning(f'[{self.name}: {current_attempt}/{auto_retry_limit}] API Error: {e}. Sleeping for 10 seconds')
            time.sleep(10.0)
            current_attempt += 1
            return self.create(
                input_object = input_object,
                parse_stream = parse_stream,
                auto_retry = auto_retry,
                auto_retry_limit = auto_retry_limit,
                _current_attempt = current_attempt,
                **kwargs
            )

        except (InvalidMaxTokens, InvalidRequestError) as e:
            raise e
        
        except Exception as e:
            if current_attempt >= auto_retry_limit:
                raise MaxRetriesExceeded(name = self.name, attempts = current_attempt, base_exception = e) from e
            logger.warning(f'[{self.name}: {current_attempt}/{auto_retry_limit}] Unknown Error: {e}. Sleeping for 10 seconds')
            time.sleep(10.0)
            current_attempt += 1
            return self.create(
                input_object = input_object,
                parse_stream = parse_stream,
                auto_retry = auto_retry,
                auto_retry_limit = auto_retry_limit,
                _current_attempt = current_attempt,
                **kwargs
            )


    async def async_create(
        self, 
        input_object: Optional[CompletionObject] = None,
        parse_stream: Optional[bool] = True,
        auto_retry: Optional[bool] = False,
        auto_retry_limit: Optional[int] = None,
        **kwargs
    ) -> CompletionResponse:
        """
        Creates a completion for the provided prompt and parameters

        Usage:

        ```python
        >>> result = await OpenAI.completions.async_create(
        >>>    prompt = 'say this is a test',
        >>>    max_tokens = 4,
        >>>    stream = True
        >>> )
        ```

        **Parameters:**

        :model (required): ID of the model to use. You can use the List models API 
        to see all of your available models,  or see our Model overview for descriptions of them.
        Default: `text-curie-001`
        
        :prompt (optional): The prompt(s) to generate completions for, encoded as 
        a string, array of strings, array of tokens, or array of token arrays.
        Note that `<|endoftext|>` is the document separator that the model sees during 
        training, so if a prompt is not specified the model will generate as if from 
        the beginning of a new document.
        Default: `<|endoftext|>`

        :suffix (optional): The suffix that comes after a completion of inserted text.
        Default: `None`

        :max_tokens (optional): The maximum number of tokens to generate in the completion.
        The token count of your prompt plus `max_tokens` cannot exceed the model's context length. 
        Most models have a context length of 2048 tokens (except for the newest models, which 
        support 4096).
        Default: `16`

        :temperature (optional): What sampling temperature to use. Higher values means 
        the model will take more risks. Try 0.9 for more creative applications, and 0 (argmax sampling) 
        for ones with a well-defined answer. We generally recommend altering this or `top_p` but not both.
        Default: `1.0`

        :top_p (optional): An alternative to sampling with `temperature`, called nucleus 
        sampling, where the model considers the results of the tokens with `top_p` probability mass. 
        So `0.1` means only  the tokens comprising the top 10% probability mass are considered.
        We generally recommend altering this or `temperature` but not both
        Default: `1.0`

        :n (optional): How many completions to generate for each prompt.
        Note: Because this parameter generates many completions, it can quickly 
        consume your token quota. Use carefully and ensure that you have reasonable 
        settings for `max_tokens` and stop.
        Default: `1`

        :stream (optional): Whether to stream back partial progress. 
        If set, tokens will be sent as data-only server-sent events as they become 
        available, with the stream terminated by a `data: [DONE]` message. This is 
        handled automatically by the Client and enables faster response processing.
        Default: `False`

        :logprobs (optional): Include the log probabilities on the `logprobs` 
        most likely tokens, as well the chosen tokens. For example, if `logprobs` is 5, 
        the API will return a list of the 5 most likely tokens. The API will always 
        return the logprob of the sampled token, so there may be up to `logprobs+1` 
        elements in the response. The maximum value for `logprobs` is 5.
        Default: `None`

        :echo (optional): Echo back the prompt in addition to the completion
        Default: `False`

        :stop (optional): Up to 4 sequences where the API will stop generating 
        further tokens. The returned text will not contain the stop sequence.
        Default: `None`

        :presence_penalty (optional): Number between `-2.0` and `2.0`. Positive values 
        penalize new tokens based on whether they appear in the text so far, increasing the 
        model's likelihood to talk about new topics
        Default: `0.0`

        :frequency_penalty (optional): Number between `-2.0` and `2.0`. Positive 
        values penalize new tokens based on their existing frequency in the text so 
        far, decreasing the model's likelihood to repeat the same line verbatim.
        Default: `0.0`

        :best_of (optional): Generates `best_of` completions server-side and returns 
        the "best" (the one with the highest log probability per token). Results cannot be streamed.
        When used with `n`, `best_of` controls the number of candidate completions and n specifies how 
        many to return – `best_of` must be greater than `n`.
        Note: Because this parameter generates many completions, it can quickly consume your token quota. 
        Use carefully and ensure that you have reasonable settings for `max_tokens` and `stop`.
        Default: `1`

        :logit_bias (optional): Modify the likelihood of specified tokens appearing in the completion.
        Accepts a json object that maps tokens (specified by their token ID in the GPT tokenizer) to an associated 
        bias value from -100 to 100. You can use this tokenizer tool (which works for both GPT-2 and GPT-3) to 
        convert text to token IDs. Mathematically, the bias is added to the logits generated by the model prior 
        to sampling. The exact effect will vary per model, but values between -1 and 1 should decrease or increase 
        likelihood of selection; values like -100 or 100 should result in a ban or exclusive selection of the 
        relevant token.
        As an example, you can pass `{"50256": -100}` to prevent the `<|endoftext|>` token from being generated.
        Default: `None`

        :user (optional): A unique identifier representing your end-user, which can help OpenAI to 
        monitor and detect abuse.
        Default: `None`

        :auto_retry (optional): Whether to automatically retry the request if it fails due to a rate limit error.

        :auto_retry_limit (optional): The maximum number of times to retry the request if it fails due to a rate limit error.

        Returns: `CompletionResult`
        """
        
        current_attempt = kwargs.pop('_current_attempt', 0)
        if not auto_retry:
            return await super().async_create(input_object = input_object, parse_stream = parse_stream,  **kwargs)

        # Handle Auto Retry Logic
        if not auto_retry_limit: auto_retry_limit = self.settings.max_retries
        try:
            return await super().async_create(input_object = input_object, parse_stream = parse_stream, **kwargs)
        except RateLimitError as e:
            if current_attempt >= auto_retry_limit:
                raise MaxRetriesExceeded(name = self.name, attempts = current_attempt, base_exception = e) from e
            sleep_interval = e.retry_after_seconds * 1.5 if e.retry_after_seconds else 15.0
            logger.warning(f'[{self.name}: {current_attempt}/{auto_retry_limit}] Rate Limit Error. Sleeping for {sleep_interval} seconds')
            await asyncio.sleep(sleep_interval)
            current_attempt += 1
            return await self.async_create(
                input_object = input_object,
                parse_stream = parse_stream,
                auto_retry = auto_retry,
                auto_retry_limit = auto_retry_limit,
                _current_attempt = current_attempt,
                **kwargs
            )
        
        except APIError as e:
            if current_attempt >= auto_retry_limit:
                raise MaxRetriesExceeded(name = self.name, attempts = current_attempt, base_exception = e) from e
            logger.warning(f'[{self.name}: {current_attempt}/{auto_retry_limit}] API Error: {e}. Sleeping for 10 seconds')
            await asyncio.sleep(10.0)
            current_attempt += 1
            return await self.async_create(
                input_object = input_object,
                parse_stream = parse_stream,
                auto_retry = auto_retry,
                auto_retry_limit = auto_retry_limit,
                _current_attempt = current_attempt,
                **kwargs
            )

        except (InvalidMaxTokens, InvalidRequestError) as e:
            raise e
        
        except Exception as e:
            if current_attempt >= auto_retry_limit:
                raise MaxRetriesExceeded(name = self.name, attempts = current_attempt, base_exception = e) from e
            logger.warning(f'[{self.name}: {current_attempt}/{auto_retry_limit}] Unknown Error: {e}. Sleeping for 10 seconds')
            await asyncio.sleep(10.0)
            current_attempt += 1
            return await self.async_create(
                input_object = input_object,
                parse_stream = parse_stream,
                auto_retry = auto_retry,
                auto_retry_limit = auto_retry_limit,
                _current_attempt = current_attempt,
                **kwargs
            )


    




## Base Object Models
from async_openai.schemas.completions import CompletionChoice, CompletionObject, CompletionResponse
from async_openai.schemas.chat import ChatMessage, ChatChoice, ChatObject, ChatResponse
from async_openai.schemas.edits import EditChoice, EditObject, EditResponse
from async_openai.schemas.embeddings import EmbeddingData, EmbeddingObject, EmbeddingResponse
# from async_openai.schemas.files import FileChoice, FileObject, FileResponse
from async_openai.schemas.images import ImageData, ImageObject, ImageResponse
from async_openai.schemas.models import ModelData, ModelObject, ModelResponse

## Route Models
from async_openai.schemas.completions import CompletionRoute
from async_openai.schemas.chat import ChatRoute
from async_openai.schemas.edits import EditRoute
from async_openai.schemas.embeddings import EmbeddingRoute
# from async_openai.schemas.files import FileRoute
from async_openai.schemas.images import ImageRoute

from async_openai.schemas.models import ModelRoute





from enum import Enum
from typing import Optional, Union
from lazyops.types import BaseModel, lazyproperty

"""
Pricing Options

# https://openai.com/api/pricing/
"""

_image_prices = {
    'small': 0.016,
    'medium': 0.018,
    'large': 0.02,
}

# price per 1k/tokens
_completion_prices = {
    'ada': 0.0004,
    'babbage': 0.0005,
    'curie': 0.002,
    'davinci': 0.02,
}

_finetune_training_prices = {
    'ada': 0.0004,
    'babbage': 0.0006,
    'curie': 0.003,
    'davinci': 0.03,
}

_finetune_usage_prices = {
    'ada': 0.0016,
    'babbage': 0.0024,
    'curie': 0.012,
    'davinci': 0.12,
}

_embedding_prices = {
    # 'ada': 0.004,
    'ada': 0.0001,
    'babbage': 0.005,
    'curie': 0.02,
    'davinci': 0.2,
}

_chat_prices = {
    'gpt-3.5-turbo': 0.002,

}

_chat_gpt_prices = {
    'gpt-4-32k': {
        'prompt': 0.06,
        'completion': 0.12,
    },
    'gpt-3.5-turbo-16k': {
        'prompt': 0.003,
        'completion': 0.004,
    },
    'gpt-35-turbo-16k': {
        'prompt': 0.003,
        'completion': 0.004,
    },
    'gpt-3-turbo-16k': {
        'prompt': 0.003,
        'completion': 0.004,
    },
    'gpt-4': {
        'prompt': 0.03,
        'completion': 0.06,
    },
    'gpt-3.5-turbo': {
        'prompt': 0.0015,
        'completion': 0.002,
    },
    'gpt-35-turbo': {
        'prompt': 0.0015,
        'completion': 0.002,
    },
    'gpt-3-turbo': {
        'prompt': 0.0015,
        'completion': 0.002,
    },
    'gpt-3.5-turbo-instruct': {
        'prompt': 0.0015,
        'completion': 0.002,
    },
    'gpt-35-turbo-instruct': {
        'prompt': 0.0015,
        'completion': 0.002,
    },
    'gpt-3-turbo-instruct': {
        'prompt': 0.0015,
        'completion': 0.002,
    },

}

# TODO rework this module

_cost_modes = {
    'embedding': _embedding_prices,
    'train': _finetune_training_prices,
    'finetune': _finetune_usage_prices,
    'completion': _completion_prices,
}

def get_arch(
    model_name: str,
) -> str:
    """
    Get the arch
    """
    for arch in {
        'babbage',
        'curie',
        'davinci',
        'ada',
    }:
        if arch in model_name:
            return arch

def get_consumption_cost(
    model_name: str,
    total_tokens: int = 1,
    default_token_cost: Optional[float] = 0.00001,
    prompt_tokens: Optional[int] = None,
    completion_tokens: Optional[int] = None,
    mode: Optional[str] = None,
) -> float:
    """
    Returns the total cost of the model
    usage
    """
    if prompt_tokens and completion_tokens:
        total_tokens = prompt_tokens + completion_tokens
    if (not mode or mode == 'chat') and any(
        arch in model_name for arch in {
            'gpt-3.5',
            'gpt-35',
            'gpt-3',
            'gpt-4',
        }):
            return next(
                (
                    (
                        prompt_tokens * (_chat_gpt_prices[gpt_model]['prompt'] / 1000)
                    )
                    + (
                        completion_tokens * (_chat_gpt_prices[gpt_model]['completion'] / 1000)
                    )
                    if prompt_tokens and completion_tokens
                    else total_tokens
                    * (
                        (
                            (_chat_gpt_prices[gpt_model]['prompt'] + _chat_gpt_prices[gpt_model]['completion']) / 2
                        )
                        / 1000
                    )
                    for gpt_model in _chat_gpt_prices
                    if gpt_model in model_name
                ),
                total_tokens * (_chat_prices['gpt-3.5-turbo'] / 1000),
            )

    arch = get_arch(model_name)
    return  total_tokens * (_cost_modes[mode][arch] / 1000) if mode in _cost_modes else total_tokens * default_token_cost



class ApiType(str, Enum):
    azure = "azure"
    openai = "openai"
    open_ai = "openai"
    azure_ad = "azure_ad"
    azuread = "azure_ad"

    def get_version(
        self, 
        version: Optional[str] = None
    ):
        if self.value in {"azure", "azure_ad", "azuread"} and not version:
            return "2023-07-01-preview"
        return version


class FilePurpose(str, Enum):
    """
    File Purpose
    """

    finetune = "fine-tune"
    fine_tune = "fine-tune"
    train = "fine-tune-train"
    search = "search"

    @classmethod
    def parse_str(cls, value: Union[str, 'FilePurpose'], raise_error: bool = True):
        if isinstance(value, cls): return value
        if "train" in value:
            return cls.train
        elif "finetune" in value:
            return cls.finetune
        elif "fine-tune" in value:
            return cls.fine_tune
        elif "search" in value:
            return cls.search
        if not raise_error: return None
        raise ValueError(f"Cannot convert {value} to FilePurpose")


class OpenAIModelType(str, Enum):
    """
    OpenAI Model Types
    """
    text = "text"
    audio = "audio"
    code = "code"
    chat = "chat"
    custom = "custom"

    @classmethod
    def parse(cls, value: Union[str, 'OpenAIModelType'], raise_error: bool = True):
        if isinstance(value, cls): return value
        if "text" in value:
            return cls.text
        elif "audio" in value:
            return cls.audio
        elif "code" in value:
            return cls.code
        elif "gpt-3.5" in value or "gpt-4" in value or "chat" in value:
            return cls.chat
        return cls.custom


class OpenAIModelArch(str, Enum):
    """
    OpenAI Model Architectures
    """

    davinci = "davinci"
    curie = "curie"
    babbage = "babbage"
    ada = "ada"
    chat = "gpt-3.5"
    chat_gpt4 = "gpt-4"
    custom = "custom"

    @classmethod
    def parse(cls, value: Union[str, 'OpenAIModelArch'], raise_error: bool = True):
        if isinstance(value, cls): return value
        if "davinci" in value:
            return cls.davinci
        elif "curie" in value:
            return cls.curie
        elif "babbage" in value:
            return cls.babbage
        elif "ada" in value:
            return cls.ada
        elif "gpt-4" in value:
            return cls.chat_gpt4
        elif "gpt-3.5" in value or "gpt-3" in value or "chat" in value:
            return cls.chat
        return cls.custom


    @lazyproperty
    def model_version(self):
        return "003" if self.value == "davinci" else "001"

    @lazyproperty
    def edit_model(self):
        return f"text-{self.value}-edit-{self.model_version}"
    
    @lazyproperty
    def completion_model(self):
        return f"text-{self.value}-{self.model_version}"
    
    @lazyproperty
    def embedding_model(self):
        return f"text-similarity-{self.value}-{self.model_version}"
    
    @lazyproperty
    def chat_model(self):
        return 'gpt-3.5-turbo' if self.value == 'gpt-3.5' else self.value
    
    @lazyproperty
    def finetune_model(self):
        return self.value
    

class ModelMode(str, Enum):
    """
    Model Mode
    """

    completion = "completion"
    edit = "edit"
    finetune = "finetune"
    fine_tune = "finetune"
    train = "train"
    embedding = "embedding"
    similiarity = "similiarity"
    search = "search"
    chat = "chat"

    @classmethod
    def parse(cls, value: Union[str, 'ModelMode'], raise_error: bool = True):
        if isinstance(value, cls): return value
        if "completion" in value:
            return cls.completion
        if "edit" in value:
            return cls.edit
        if "finetune" in value:
            return cls.finetune
        if "fine-tune" in value:
            return cls.fine_tune
        if "train" in value:
            return cls.train
        if "embedding" in value:
            return cls.embedding
        if "search" in value:
            return cls.search
        if "similiarity" in value:
            return cls.similiarity
        if "gpt-3.5" in value or "gpt-35" in value or 'gpt-4' in value or "chat" in value:
            return cls.chat
        if "text" in value:
            return cls.completion
        if not raise_error: return None
        raise ValueError(f"Cannot convert {value} to ModelMode")
    
    @classmethod
    def get_text_modes(cls):
        return [
            cls.completion, 
            cls.edit,
            cls.embedding,
            cls.similiarity, 
            cls.search, 
            # cls.chat
        ]

class OpenAIModel(object):

    def __init__(
        self, 
        value: str, 
        **kwargs
    ):
        self.src_value = value
        self.src_splits = value.split("-")
        self.mode: ModelMode = kwargs.get("mode")
        self.model_arch: OpenAIModelArch = kwargs.get("model_arch")
        self.model_type: OpenAIModelType = kwargs.get("model_type")
        self.version: str = kwargs.get("version")
        self.parse_values()

    def parse_values(self):
        """
        Parse the source values into the correct parts
        """
        self.mode = ModelMode.parse((self.mode or self.src_value), raise_error = False) or ModelMode.completion
        self.model_arch = OpenAIModelArch.parse((self.model_arch or self.src_value), raise_error = False)
        self.model_type = OpenAIModelType.parse(
            (self.model_type or \
                ("text" if self.mode in ModelMode.get_text_modes() else self.src_value)
            ), raise_error = False)
        if not self.version:
            ver_values = [x for x in self.src_splits if (x[0].isdigit() and x[-1].isdigit())]
            if ver_values:
                self.version = '-'.join(ver_values)
                if self.mode in {ModelMode.chat}:
                    if self.version in {'35', '3.5', '4', '3', '16k', '32k'}:
                        self.version = None
                    else:
                        self.version = self.version.rsplit('-', 1)[-1]

            elif self.mode == ModelMode.completion:
                self.version = "003" if self.model_arch == "davinci" else "001"
            elif self.mode == ModelMode.chat:
                pass
            elif self.model_type != OpenAIModelType.custom:
                self.version = "001"


    @lazyproperty
    def value(self) -> str:
        """
        The value of the model
        """
        if self.model_arch == OpenAIModelArch.custom or self.model_type == OpenAIModelType.custom:
            return self.src_value
        if self.mode == ModelMode.chat:
            return f'{self.src_value}-{self.version}' if self.version else self.src_value
        t = f'{self.model_type.value}'
        if self.mode != ModelMode.completion:
            t += f'-{self.mode.value}'
        t += f'-{self.model_arch.value}'
        if self.version:
            t += f'-{self.version}'
        return t

    def dict(self, *args, **kwargs):
        return {
            "value": self.value,
            "mode": self.mode.value,
            "model_arch": self.model_arch.value,
            "model_type": self.model_type.value,
            "version": self.version,
        }
    
    def __str__(self):
        return f'OpenAIModel(value="{self.value}", mode="{self.mode}", model_arch="{self.model_arch}", model_type="{self.model_type}", version="{self.version}")'

    def __repr__(self) -> str:
        return f'OpenAIModel(value="{self.value}", mode="{self.mode}", model_arch="{self.model_arch}", model_type="{self.model_type}", version="{self.version}")'

    def __json__(self):
        return self.value
    
    # def __dict__(self):
    #     return self.value

    def get_cost(
        self,
        total_tokens: int = 1,
        mode: Optional[str] = None,
        raise_error: bool = True,
        default_token_cost: Optional[float] = 0.00001,
        prompt_tokens: Optional[int] = None,
        completion_tokens: Optional[int] = None,
    ) -> float:
        """
        Returns the total cost of the model
        usage
        """
        if prompt_tokens and completion_tokens:
            total_tokens = prompt_tokens + completion_tokens

        mode = mode or self.mode.value
        if mode in ['completion', 'edit']:
            return total_tokens * (_completion_prices[self.model_arch.value] / 1000)
        if mode in ['chat']:
            return next(
                (
                    (
                        prompt_tokens * (_chat_gpt_prices[gpt_model]['prompt'] / 1000)
                    )
                    + (
                        completion_tokens * (_chat_gpt_prices[gpt_model]['completion'] / 1000)
                    )
                    if prompt_tokens and completion_tokens
                    else total_tokens
                    * (
                        (
                            (_chat_gpt_prices[gpt_model]['prompt'] + _chat_gpt_prices[gpt_model]['completion']) / 2
                        )
                        / 1000
                    )
                    for gpt_model in _chat_gpt_prices
                    if gpt_model in self.src_value
                ),
                total_tokens * (_chat_prices['gpt-3.5-turbo'] / 1000),
            )
        if 'embedding' in mode:
            return total_tokens * (_embedding_prices[self.model_arch.value]  / 1000)
        if 'train' in mode:
            return total_tokens * (_finetune_training_prices[self.model_arch.value] / 1000)
        if 'finetune' in mode or 'fine-tune' in mode:
            return total_tokens * (_finetune_usage_prices[self.model_arch.value] / 1000)
        if raise_error: raise ValueError(f"Invalid mode {mode}")
        return total_tokens * default_token_cost


class EditModels(str, Enum):
    """
    Just the base models available
    """
    davinci = "text-davinci-edit-003"
    curie = "text-curie-edit-001"
    babbage = "text-babbage-edit-001"
    ada = "text-ada-edit-001"

    @lazyproperty
    def model_type(self) -> str:
        return self.value.split("-")[1]

class EmbeddingModels(str, Enum):
    """
    Just the base models available
    """
    davinci = "text-similarity-davinci-003"
    curie = "text-similarity-curie-001"
    babbage = "text-similarity-babbage-001"
    ada = "text-similarity-ada-001"

    @lazyproperty
    def model_type(self) -> str:
        return self.value.split("-")[2]

class CompletionModels(str, Enum):
    """
    Just the base models available
    """
    davinci = "text-davinci-003"
    curie = "text-curie-001"
    babbage = "text-babbage-001"
    ada = "text-ada-001"

    @lazyproperty
    def model_type(self) -> str:
        return self.value.split("-")[1]


class FinetuneModels(str, Enum):
    """
    Supported finetune models.
    """
    ada = "ada"
    babbage = "babbage"
    curie = "curie"
    davici = "davici"

    @lazyproperty
    def model_type(self):
        return self.value

class ImageSize(str, Enum):
    """
    Size of the image
    """

    small = "256x256"
    medium = "512x512"
    large = "1024x1024"

    @lazyproperty
    def image_type(self):
        if self.value == "256x256":
            return 'small'
        if self.value == "512x512":
            return 'medium'
        if self.value == "1024x1024":
            return 'large'
        raise ValueError(f"Cannot convert {self.value} to Kind")

    @classmethod
    def from_str(cls, value: str) -> "ImageSize":
        """
        :param value: Size of the image
        :type value: str
        :return: ImageSize
        :rtype: ImageSize
        """
        if value == "256x256":
            return cls.small
        if value == "512x512":
            return cls.medium
        if value == "1024x1024":
            return cls.large
        try:
            return cls(value)
        except ValueError as e:
            raise ValueError(f"Cannot convert {value} to ImageSize") from e
    
    def get_cost(
        self,
        total_images: int = 1,
    ) -> float:
        """
        Returns the total cost of the model
        usage
        """
        return total_images * _image_prices[self.value]

class ImageFormat(str, Enum):
    """
    Format of the image
    """

    url = "url"
    b64 = "b64_json"
    b64_json = "b64_json"


from __future__ import absolute_import

from async_openai.types.errors import (
    OpenAIError,
    APIError,
    AuthenticationError,
    InvalidRequestError,
    RateLimitError,
    APIConnectionError,
    Timeout,
    TryAgain,
    ServiceUnavailableError,
    fatal_exception,
    error_handler,
)

from async_openai.types.options import (
    ApiType,
    CompletionModels,
    FilePurpose,
    FinetuneModels,
    ImageSize,
    ImageFormat,
)

# from async_openai.types.base import (
#     Usage,
#     Permission,
#     BaseResource,
#     FileObject,
#     EventObject,
#     FileResource,
#     BaseResponse,
#     BaseRoute,

#     RESPONSE_SUCCESS_CODES
# )




from __future__ import annotations

"""
OpenAI Client that supports external providers and configurations
that have OpenAI-compatible endpoints.
"""

import abc
import aiohttpx
import contextlib
from typing import Optional, Callable, Dict, Union, List

from async_openai.schemas import *
from async_openai.utils.config import get_settings, OpenAISettings
from async_openai.utils.logs import logger
from async_openai.utils.config import ProxyObject
from async_openai.utils.external_config import ExternalProviderSettings, ExternalProviderAuth
from async_openai.routes import ApiRoutes


class ExternalOpenAIClient(abc.ABC):
    """
    External Client for all the routes in the API.
    """

    _client: Optional[aiohttpx.Client] = None
    _routes: Optional[ApiRoutes] = None
    _kwargs: Optional[Dict] = None

    def __init__(
        self,
        name: str,
        provider: ExternalProviderSettings,
        is_proxied: Optional[bool] = None,
        **kwargs
    ):  
        """
        Lazily Instantiates the OpenAI Client
        """
        self.name = name
        self.provider = provider
        self.debug_enabled: Optional[bool] = None
        self.on_error: Optional[Callable] = None
        self.timeout: Optional[int] = None
        self.max_retries: Optional[int] = None
        self.ignore_errors: Optional[bool] = None
        self.disable_retries: Optional[bool] = None
        self.retry_function: Optional[Callable] = None

        self.is_proxied = is_proxied if is_proxied is not None else \
           (self.provider.config.has_proxy and  '_noproxy' not in self.name)
        
        self.settings: Optional[OpenAISettings] = kwargs.pop('settings', get_settings())
        self.client_callbacks: List[Callable] = []
        self.auth: Optional[ExternalProviderAuth] = None
        self.configure_params(**kwargs)


    @property
    def client(self) -> aiohttpx.Client:
        """
        Returns the aiohttpx client
        """
        if self._client is None:
            self.configure_client()
        return self._client
    
    @property
    def routes(self) -> ApiRoutes:
        """
        Returns the routes class
        """
        if self._routes is None:
            self.configure_routes()
        return self._routes

    def configure_params(
        self, 
        debug_enabled: Optional[bool] = None,
        on_error: Optional[Callable] = None,
        timeout: Optional[int] = None,
        max_retries: Optional[int] = None,
        ignore_errors: Optional[bool] = None,
        disable_retries: Optional[bool] = None,
        retry_function: Optional[Callable] = None,
        auth: Optional[ExternalProviderAuth] = None,
        client_callbacks: Optional[List[Callable]] = None,
        **kwargs
    ):  # sourcery skip: low-code-quality
        """
        Helper to configure the client
        """
        
        if debug_enabled is not None:
            self.debug_enabled = debug_enabled
        elif self.debug_enabled is None:
            self.debug_enabled = self.settings.debug_enabled
        
        if timeout is not None:
            self.timeout = timeout
        elif self.timeout is None:
            self.timeout = self.settings.timeout

        if on_error is not None:
            self.on_error = on_error
        if ignore_errors is not None:
            self.ignore_errors = ignore_errors
        elif self.ignore_errors is None:
            self.ignore_errors = self.settings.ignore_errors
        if max_retries is not None:
            self.max_retries = max_retries
        elif self.max_retries is None:
            self.max_retries = self.settings.max_retries
        if disable_retries is not None:
            self.disable_retries = disable_retries
        elif self.disable_retries is None:
            self.disable_retries = self.settings.disable_retries
        if retry_function is not None:
            self.retry_function = retry_function
        
        if auth is not None:
            self.auth = auth
        elif self.auth is None:
            self.auth = ExternalProviderAuth(config = self.provider.config, is_proxied = self.is_proxied)
        
        if kwargs: self._kwargs = kwargs
        self.log_method = logger.info if self.debug_enabled else logger.debug
        if not self.debug_enabled:
            self.settings.disable_httpx_logger()
        
        if client_callbacks is not None:
            self.client_callbacks = client_callbacks

    def configure_client(self, **kwargs):
        """
        Helper to configure the client
        """
        if self._client is not None: return
        # logger.info(f"OpenAI Client Configured: {self.base_url} [{self.name}]")
        extra_kwargs = {}
        
        self._client = aiohttpx.Client(
            base_url = self.provider.config.proxy_url if self.is_proxied else self.provider.config.api_url,
            timeout = self.timeout,
            limits = self.settings.api_client_limits,
            auth = self.auth,
            headers = {
                'content-type': 'application/json',
            },
            **extra_kwargs,
        )

    def configure_routes(self, **kwargs):
        """
        Helper to configure the client routes
        """
        if self._routes is not None: return
        kwargs = kwargs or {}
        if self._kwargs: kwargs.update(self._kwargs)
        self._routes = ApiRoutes(
            client = self.client,
            name = self.provider.name,
            # headers = self.headers,
            debug_enabled = self.debug_enabled,
            on_error = self.on_error,
            ignore_errors = self.ignore_errors,
            timeout = self.timeout,
            max_retries = self.max_retries,
            settings = self.settings,
            disable_retries = self.disable_retries,
            retry_function = self.retry_function,
            client_callbacks = self.client_callbacks,
            route_classes = self.provider.routes.api_route_classes,
            is_azure = False,
            **kwargs
        )
        if self.debug_enabled:
            logger.info(f"[{self.name}] External Provider Configured: {self.client.base_url}")
            logger.debug(f"Debug Enabled: {self.debug_enabled}")


    def reset(
        self,
        **kwargs
    ):
        """
        Resets the client to the default settings
        """
        self._client = None
        self._routes = None
        self.configure_params(**kwargs)


    @property
    def completions(self) -> CompletionRoute:
        """
        Returns the `CompletionRoute` class for interacting with `Completions`.
        
        Doc: `https://beta.openai.com/docs/api-reference/completions`
        """
        return self.routes.completions
    
    @property
    def chat(self) -> ChatRoute:
        """
        Returns the `ChatRoute` class for interacting with `Chat` components

        Doc: `https://platform.openai.com/docs/api-reference/chat`
        """
        return self.routes.chat

    @property
    def edits(self) -> EditRoute:
        """
        Returns the `EditRoute` class for interacting with `Edits`.
        
        Doc: `https://beta.openai.com/docs/api-reference/edits`
        """
        return self.routes.edits
    
    @property
    def embeddings(self) -> EmbeddingRoute:
        """
        Returns the `EmbeddingRoute` class for interacting with `Embeddings`.
        
        Doc: `https://beta.openai.com/docs/api-reference/embeddings`
        """
        return self.routes.embeddings
    
    @property
    def images(self) -> ImageRoute:
        """
        Returns the `ImageRoute` class for interacting with `Images`.
        
        Doc: `https://beta.openai.com/docs/api-reference/images`
        """
        return self.routes.images
    
    @property
    def models(self) -> ModelRoute:
        """
        Returns the `ModelRoute` class for interacting with `models`.
        
        Doc: `https://beta.openai.com/docs/api-reference/models`
        """
        return self.routes.models

    """
    Context Managers
    """

    async def async_close(self):
        await self.client.aclose()
    
    def close(self):
        self.client.close()
    
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_value, traceback):
        self.close()
    
    async def __aenter__(self):
        return self
    
    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.async_close()


    def ping(self, timeout: Optional[float] = 1.0) -> bool:
        """
        Pings the API Endpoint to check if it's alive.
        """
        try:
        # with contextlib.suppress(Exception):
            response = self.client.get('/', timeout = timeout)
            data = response.json()
            # we should expect a 404 with a json response
            # if self.debug_enabled: logger.info(f"API Ping: {data}\n{response.headers}")
            if data.get('error'): return True
        except Exception as e:
            logger.error(f"API Ping Failed: {e}")
        return False
    
    async def aping(self, timeout: Optional[float] = 1.0) -> bool:
        """
        Pings the API Endpoint to check if it's alive.
        """
        with contextlib.suppress(Exception):
            response = await self.client.async_get('/', timeout = timeout)
            data = response.json()
            # we should expect a 404 with a json response
            if data.get('error'): return True
        return False





from __future__ import annotations

"""
OpenAI Functions Base Class
"""

import jinja2
import functools
import inspect
from abc import ABC
from pydantic import Field, BaseModel
# from lazyops.types import BaseModel
from lazyops.utils.times import Timer
from lazyops.libs.proxyobj import ProxyObject
from lazyops.types.models import schema_extra, PYD_VERSION
from async_openai.utils.fixjson import resolve_json
from . import errors

from typing import Optional, Any, Set, Dict, List, Union, Type, Tuple, Awaitable, Generator, AsyncGenerator, TypeVar, TYPE_CHECKING

if PYD_VERSION == 2:
    from pydantic import ConfigDict

if TYPE_CHECKING:
    from async_openai import ChatResponse, ChatRoute
    from async_openai.types.resources import Usage
    from async_openai.manager import OpenAIManager as OpenAISessionManager
    from lazyops.utils.logs import Logger
    from lazyops.libs.persistence import PersistentDict
    

FT = TypeVar('FT', bound = BaseModel)
SchemaT = TypeVar('SchemaT', bound = BaseModel)


class BaseFunctionModel(BaseModel):
    function_name: Optional[str] = Field(None, hidden = True)
    function_model: Optional[str] = Field(None, hidden = True)
    function_duration: Optional[float] = Field(None, hidden = True)

    if TYPE_CHECKING:
        function_usage: Optional[Usage]
    else:
        function_usage: Optional[Any] = Field(None, hidden = True)

    def update(
        self,
        values: 'BaseFunctionModel',
    ):
        """
        Updates the values
        """
        pass

    def _setup_item(
        self,
        item: 'SchemaT',
        **kwargs
    ) -> 'SchemaT':
        """
        Updates the Reference Item
        """
        return item
    

    def update_values(
        self,
        item: 'SchemaT',
        **kwargs
    ) -> 'SchemaT':
        """
        Updates the Reference Item with the values
        """
        return item
    

    def update_data(
        self,
        item: 'SchemaT',
        **kwargs
    ) -> 'SchemaT':
        """
        Updates the Reference Item with the values
        """
        item = self._setup_item(item = item, **kwargs)
        item = self.update_values(item = item, **kwargs)
        return item
    

    def is_valid(self) -> bool:
        """
        Returns whether the function data is valid
        """
        return True
    
    def _set_values_from_response(
        self,
        response: 'ChatResponse',
        name: Optional[str] = None,
        **kwargs
    ) -> 'BaseFunctionModel':
        """
        Sets the values from the response
        """
        if name: self.function_name = name
        self.function_usage = response.usage
        if response.response_ms: self.function_duration = response.response_ms / 1000
        self.function_model = response.model

    @property
    def function_cost(self) -> Optional[float]:
        """
        Returns the function consumption
        """
        if not self.function_model: return None
        if not self.function_usage: return None
        from async_openai.types.context import ModelContextHandler
        return ModelContextHandler.get_consumption_cost(self.function_model, self.function_usage)
    
    @property
    def function_cost_string(self) -> Optional[str]:
        """
        Returns the function consumption as a pretty string
        """
        return f"${self.function_cost:.2f}" if self.function_cost else None


    if PYD_VERSION == 2:
        model_config = ConfigDict(json_schema_extra = schema_extra, arbitrary_types_allowed = True)
    else:
        class Config:
            json_schema_extra = schema_extra
            arbitrary_types_allowed = True



FunctionSchemaT = TypeVar('FunctionSchemaT', bound = BaseFunctionModel)
FunctionResultT = TypeVar('FunctionResultT', bound = BaseFunctionModel)

class BaseFunction(ABC):
    """
    Base Class for OpenAI Functions
    """
    
    name: Optional[str] = None
    function_name: Optional[str] = None
    description: Optional[str] = None
    schema: Optional[Type[FunctionSchemaT]] = None
    schemas: Optional[Dict[str, Dict[str, Union[str, Type[FunctionSchemaT]]]]] = None
    
    prompt_template: Optional[str] = None
    system_template: Optional[str] = None

    default_model: Optional[str] = 'gpt-35-turbo'
    default_larger_model: Optional[bool] = None
    cachable: Optional[bool] = True
    result_buffer: Optional[int] = 1000
    retry_limit: Optional[int] = 5
    max_attempts: Optional[int] = 2

    default_model_local: Optional[str] = None
    default_model_develop: Optional[str] = None
    default_model_production: Optional[str] = None

    auto_register_function: Optional[bool] = True

    def __init_subclass__(cls, **kwargs):
        """
        Subclass Hook
        """
        if cls.auto_register_function:
            OpenAIFunctions.register_function(cls, initialize = False)


    def __init__(
        self, 
        api: Optional['OpenAISessionManager'] = None,
        debug_enabled: Optional[bool] = None,
        **kwargs
    ):
        """
        This gets initialized from the Enrichment Handler
        """
        from async_openai.manager import ModelContextHandler
        from async_openai.utils.logs import logger, null_logger
        self.ctx: Type['ModelContextHandler'] = ModelContextHandler
        if api is None:
            from async_openai.client import OpenAIManager
            api = OpenAIManager
        
        self.api: 'OpenAISessionManager' = api
        self.pool = self.api.pooler
        self.kwargs = kwargs
        self.logger = logger
        self.null_logger = null_logger
        self.settings = self.api.settings
        if debug_enabled is not None:
            self.debug_enabled = debug_enabled
        else:
            self.debug_enabled = self.settings.debug_enabled
        self.build_funcs(**kwargs)
        self.build_templates(**kwargs)
        self.post_init(**kwargs)

    @property
    def default_model_func(self) -> str:
        """
        Returns the default model
        """
        if self.settings.is_local_env:
            return self.default_model_local or self.default_model
        if self.settings.is_development_env:
            return self.default_model_develop or self.default_model
        return self.default_model_production or self.default_model

    @property
    def autologger(self) -> 'Logger':
        """
        Returns the logger
        """
        return self.logger if \
            (self.debug_enabled or self.settings.is_development_env) else self.null_logger


    @property
    def has_diff_model_than_default(self) -> bool:
        """
        Returns True if the default model is different than the default model
        """
        return self.default_model_func != self.default_model


    def build_templates(self, **kwargs):
        """
        Construct the templates
        """
        self.template = self.create_template(self.prompt_template)
        # Only create the system template if it's a jinja template
        if self.system_template and '{%' in self.system_template:
            self.system_template = self.create_template(self.system_template)

    def build_funcs(self, **kwargs):
        """
        Builds the functions
        """
        # Handles multi functions
        if self.schemas:
            self.functions = []
            self.functions.extend(
                {
                    "name": name,
                    "description": data.get('description', self.description),
                    "parameters": data.get('schema', self.schema),
                }
                for name, data in self.schemas.items()
            )
        else:
            self.functions = [
                {
                    "name": self.function_name or self.name,
                    "description": self.description,
                    "parameters": self.schema,
                }
            ]


    def post_init(self, **kwargs):
        """
        Post Init Hook
        """
        pass
    
    def pre_call_hook(self, *args, **kwargs):
        """
        Pre Call Hook
        """
        pass

    async def apre_call_hook(self, *args, **kwargs):
        """
        Pre Call Hook
        """
        self.pre_call_hook(*args, **kwargs)

    def pre_validate(self, *args, **kwargs) -> bool:
        """
        Validate the input before running
        """
        return True

    async def apre_validate(self, *args, **kwargs) -> bool:
        """
        Validate the input before running
        """
        return self.pre_validate(*args, **kwargs)
    

    def pre_validate_model(self, prompt: str, model: str, *args, **kwargs) -> str:
        """
        Validates the model before running
        """
        return model

    async def apre_validate_model(self, prompt: str, model: str, *args, **kwargs) -> str:
        """
        Validates the model before running
        """
        return self.pre_validate_model(prompt = prompt, model = model, *args, **kwargs)
    


    def call(
        self,
        *args,
        model: Optional[str] = None, 
        **kwargs
    ) -> Optional[FunctionSchemaT]:
        """
        Call the function
        """
        if not self.pre_validate(*args, **kwargs):
            return None
        self.pre_call_hook(*args, **kwargs)
        return self.run_function(*args, model = model, **kwargs)
    
    async def acall(
        self,
        *args,
        model: Optional[str] = None, 
        **kwargs
    ) -> Optional[FunctionSchemaT]:
        """
        Call the function
        """
        if not await self.apre_validate(*args, **kwargs):
            return None
        await self.apre_call_hook(*args, **kwargs)
        return await self.arun_function(*args, model = model, **kwargs)

    def __call__(
        self,
        *args,
        model: Optional[str] = None, 
        is_async: Optional[bool] = True,
        **kwargs
    ) -> Optional[FunctionSchemaT]:
        """
        Call the function
        """
        if is_async: return self.acall(*args, model = model, **kwargs)
        return self.call(*args, model = model, **kwargs)

    def get_chat_client(self, model: str, **kwargs) -> 'ChatRoute':
        """
        Gets the chat client
        """
        return self.api.get_chat_client(model = model, **kwargs)
    
    def get_completion_client(self, model: str, **kwargs) -> 'ChatRoute':
        """
        Gets the chat client
        """
        return self.api.get_chat_client(model = model, **kwargs)
    
    async def arun_chat_function(
        self, 
        messages: List[Dict[str, Any]],
        chat: Optional['ChatRoute'] = None, 
        cachable: Optional[bool] = None,
        functions: Optional[List[Dict[str, Any]]] = None,
        function_name: Optional[str] = None,
        property_meta: Optional[Dict[str, Any]] = None,
        model: Optional[str] = None,
        headers: Optional[Dict[str, str]] = None,
        excluded_clients: Optional[List[str]] = None,
        **kwargs,
    ) -> ChatResponse:  # sourcery skip: low-code-quality
        """
        Runs the chat function
        """
        current_attempt = kwargs.pop('_current_attempt', 0)
        last_chat_name = kwargs.pop('_last_chat_name', None)
        if current_attempt and current_attempt > self.retry_limit:
            raise errors.MaxRetriesExhausted(
                name = last_chat_name,
                func_name = function_name or self.name,
                model = model,
                attempts = current_attempt,
                max_attempts = self.retry_limit,
            )

        disable_cache = not cachable if cachable is not None else not self.cachable
        if not chat: 
            if last_chat_name:
                if not excluded_clients: excluded_clients = []
                excluded_clients.append(last_chat_name)
            chat = self.get_chat_client(model = model, excluded_clients = excluded_clients, **kwargs)
        if not headers and 'noproxy' not in chat.name:
            headers = {
                'Helicone-Cache-Enabled': 'false' if disable_cache else 'true',
                'Helicone-Property-FunctionName': function_name or self.name,
            }
            if property_meta:
                property_meta = {f'Helicone-Property-{k}': str(v) for k, v in property_meta.items()}
                headers.update(property_meta)
            
        elif headers and 'noproxy' in chat.name:
            headers = None
        functions = functions or self.functions
        try:
            if headers: chat.client.headers.update(headers)
            return await chat.async_create(
                model = model,
                messages = messages,
                functions = functions,
                headers = headers,
                auto_retry = True,
                auto_retry_limit = 2,
                function_call = {'name': function_name or self.name},
                header_cache_keys = ['Helicone-Cache-Enabled'],
                **kwargs,
            )
        except errors.InvalidRequestError as e:
            self.logger.info(f"[{current_attempt}/{self.retry_limit}] [{self.name} - {model}] Invalid Request Error. |r|{e}|e|", colored=True)
            raise e
        except errors.MaxRetriesExceeded as e:
            self.autologger.info(f"[{current_attempt}/{self.retry_limit}] [{self.name} - {model}] Retrying...", colored=True)
            return await self.arun_chat_function(
                messages = messages,
                cachable = cachable,
                functions = functions,
                function_name = function_name,
                property_meta = property_meta,
                model = model,
                headers = headers,
                excluded_clients = excluded_clients,
                _current_attempt = current_attempt + 1,
                _last_chat_name = chat.name,
                **kwargs,
            )
        except Exception as e:
            self.autologger.info(f"[{current_attempt}/{self.retry_limit}] [{self.name} - {model}] Unknown Error Trying to run chat function: |r|{e}|e|", colored=True)
            return await self.arun_chat_function(
                messages = messages,
                cachable = cachable,
                functions = functions,
                function_name = function_name,
                property_meta = property_meta,
                model = model,
                headers = headers,
                excluded_clients = excluded_clients,
                _current_attempt = current_attempt + 1,
                _last_chat_name = chat.name,
                **kwargs,
            )

    def run_chat_function(
        self, 
        chat: 'ChatRoute', 
        messages: List[Dict[str, Any]],
        cachable: Optional[bool] = None,
        functions: Optional[List[Dict[str, Any]]] = None,
        function_name: Optional[str] = None,
        property_meta: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> ChatResponse:
        """
        Runs the chat function
        """
        disable_cache = not cachable if cachable is not None else not self.cachable
        headers = None
        if 'noproxy' not in chat.name:
            headers = {
                'Helicone-Cache-Enabled': 'false' if disable_cache else 'true',
                'Helicone-Property-FunctionName': function_name or self.name,
            }
            if property_meta:
                property_meta = {f'Helicone-Property-{k.strip()}': str(v).strip() for k, v in property_meta.items()}
                headers.update(property_meta)
        if headers: chat.client.headers.update(headers)
        functions = functions or self.functions
        return chat.create(
            messages = messages,
            functions = functions,
            headers = headers,
            auto_retry = True,
            auto_retry_limit = self.retry_limit,
            function_call = {'name': function_name or self.name},
            header_cache_keys=['Helicone-Cache-Enabled'],
            **kwargs,
        )

    def parse_response(
        self,
        response: 'ChatResponse',
        schema: Optional[Type[FunctionSchemaT]] = None,
        include_name: Optional[bool] = True,
    ) -> Optional[FunctionSchemaT]:  # sourcery skip: extract-duplicate-method
        """
        Parses the response
        """
        schema = schema or self.schema
        try:
            result = schema.model_validate(response.function_results[0].arguments, from_attributes = True)
            result._set_values_from_response(response, name = self.name if include_name else None)
            return result
        except Exception as e:
            self.autologger.error(f"[{self.name} - {response.model} - {response.usage}] Failed to parse object: {e}\n{response.text}\n{response.function_results[0].arguments}")
            try:
                result = schema.model_validate(resolve_json(response.function_results[0].arguments), from_attributes = True)
                result._set_values_from_response(response, name = self.name if include_name else None)
                return result
            except Exception as e:
                self.autologger.error(f"[{self.name} - {response.model} - {response.usage}] Failed to parse object after fixing")
                return None
    

    def is_valid_response(self, response: FT) -> bool:
        """
        Returns True if the response is valid
        """
        return True
    
    def apply_text_cleaning(self, text: str) -> str:
        """
        Applies text cleaning
        """
        from lazyops.utils.format_utils import clean_html, clean_text, cleanup_dots
        if "..." in text: text = cleanup_dots(text)
        return clean_html(clean_text(text))
    
    @staticmethod
    def create_template(template: str, enable_async: Optional[bool] = False, **kwargs) -> jinja2.Template:
        """
        Creates the template
        """
        return jinja2.Template(template, enable_async = enable_async, **kwargs)
    
    def truncate_documents(
        self, 
        documents: Dict[str, str],
        max_length: Optional[int] = None,
        buffer_size: Optional[int] = None,
        model: Optional[str] = None,
        truncation_length: Optional[int] = None,
    ) -> Dict[str, str]:
        """
        Helper Function to truncate supporting docs
        """
        current_length = 0
        if max_length is None:
            model = model or self.default_model_func
            max_length = self.ctx.get(model).context_length
        if buffer_size is None: buffer_size = self.result_buffer
        max_length -= buffer_size

        truncation_length = truncation_length or (max_length // len(documents))
        new_documents = {}
        for file_name, file_text in documents.items():
            if not file_text: continue
            file_text = self.apply_text_cleaning(file_text)[:truncation_length]
            current_length += len(file_text)
            new_documents[file_name] = file_text
            if current_length > max_length: break
        return new_documents
    
    """
    Function Handlers
    """

    def prepare_function_inputs(
        self,
        model: Optional[str] = None,
        **kwargs
    ) -> Tuple[List[Dict[str, Any]], str]:
        """
        Prepare the Function Inputs for the function
        """
        model = model or self.default_model_func
        prompt = self.template.render(**kwargs)
        prompt = self.api.truncate_to_max_length(prompt, model = model, buffer_length = self.result_buffer)
        messages = []
        if self.system_template:
            if isinstance(self.system_template, jinja2.Template):
                system_template = self.system_template.render(**kwargs)
            else:
                system_template = self.system_template
            messages.append({
                "role": "system",
                "content": system_template,
            })
        messages.append({
            "role": "user", 
            "content": prompt,
        })
        return messages, model
    
    async def aprepare_function_inputs(
        self,
        model: Optional[str] = None,
        **kwargs
    ) -> Tuple[List[Dict[str, Any]], str]:
        """
        Prepare the Function Inputs for the function
        """
        model = model or self.default_model_func
        prompt = self.template.render(**kwargs)
        prompt = await self.api.atruncate_to_max_length(prompt, model = model, buffer_length = self.result_buffer)
        messages = []
        if self.system_template:
            if isinstance(self.system_template, jinja2.Template):
                system_template = self.system_template.render(**kwargs)
            else:
                system_template = self.system_template
            messages.append({
                "role": "system",
                "content": system_template,
            })
        messages.append({
            "role": "user", 
            "content": prompt,
        })
        return messages, model

    def run_function(
        self,
        *args,
        model: Optional[str] = None, 
        **kwargs
    ) -> Optional[FunctionSchemaT]:
        """
        Returns the Function Result
        """
        messages, model = self.prepare_function_inputs(model = model, **kwargs)
        return self.run_function_loop(messages = messages, model = model, **kwargs)

    async def arun_function(
        self,
        *args,
        model: Optional[str] = None, 
        **kwargs
    ) -> Optional[FunctionSchemaT]:
        """
        Returns the Function Result
        """
        messages, model = await self.aprepare_function_inputs(model = model, **kwargs)
        return await self.arun_function_loop(messages = messages, model = model, **kwargs)
    
    def get_function_kwargs(self) -> Dict[str, Any]:
        """
        Returns the function kwargs
        """
        sig = inspect.signature(self.arun_function)
        return [p.name for p in sig.parameters.values() if p.kind in {p.KEYWORD_ONLY, p.VAR_KEYWORD, p.POSITIONAL_OR_KEYWORD} and p.name not in {'kwargs', 'args', 'model'}]

    """
    Handle a Loop
    """

    def run_function_loop(
        self,
        messages: List[Dict[str, Any]],
        model: Optional[str] = None,
        raise_errors: Optional[bool] = True,
        **kwargs,
    ) -> Optional[FunctionSchemaT]:
        """
        Runs the function loop
        """
        chat = self.get_chat_client(model = model, **kwargs)
        response = self.run_chat_function(
            chat = chat,
            messages = messages,
            model = model,
            **kwargs,
        )

        result = self.parse_response(response, include_name = True)
        if result is not None: return result

        # Try Again
        attempts = 1
        _ = kwargs.pop('cachable', None)
        while attempts < self.max_attempts:
            chat = self.get_chat_client(model = model, **kwargs)
            response = self.run_chat_function(
                chat = chat,
                messages = messages,
                model = model,
                **kwargs,
            )
            result = self.parse_response(response, include_name = True)
            if result is not None: return result
            attempts += 1
        self.autologger.error(f"Unable to parse the response for {self.name} after {self.max_attempts} attempts.")
        if raise_errors: raise errors.MaxRetriesExhausted(
            name = self.name, 
            func_name = self.name,
            model = model,
            attempts = self.max_attempts,
        )
        return None

    async def arun_function_loop(
        self,
        messages: List[Dict[str, Any]],
        model: Optional[str] = None,
        raise_errors: Optional[bool] = True,
        **kwargs,
    ) -> Optional[FunctionSchemaT]:
        """
        Runs the function loop
        """
        chat = self.get_chat_client(model = model, **kwargs)
        response = await self.arun_chat_function(
            chat = chat,
            messages = messages,
            model = model,
            **kwargs,
        )

        result = self.parse_response(response, include_name = True)
        if result is not None: return result

        # Try Again
        attempts = 1
        _ = kwargs.pop('cachable', None)
        while attempts < self.max_attempts:
            chat = self.get_chat_client(model = model, **kwargs)
            response = await self.arun_chat_function(
                chat = chat,
                messages = messages,
                model = model,
                **kwargs,
            )
            result = self.parse_response(response, include_name = True)
            if result is not None: return result
            attempts += 1
        self.autologger.error(f"Unable to parse the response for {self.name} after {self.max_attempts} attempts.")
        if raise_errors: raise errors.MaxRetriesExhausted(
            name = self.name, 
            func_name = self.name,
            model = model,
            attempts = self.max_attempts,
        )
        return None

    

FunctionT = TypeVar('FunctionT', bound = BaseFunction)



class FunctionManager(ABC):
    """
    The Functions Manager Class that handles registering and managing functions

    - Additionally supports caching through `kvdb`
    """

    name: Optional[str] = 'functions'

    def __init__(
        self,
        **kwargs,
    ):
        from async_openai.utils.config import settings
        from async_openai.utils.logs import logger, null_logger
        self.logger = logger
        self.null_logger = null_logger
        self.settings = settings
        self.debug_enabled = self.settings.debug_enabled
        self.cache_enabled = self.settings.function_cache_enabled

        self._api: Optional['OpenAISessionManager'] = None
        self._cache: Optional['PersistentDict'] = None
        self.functions: Dict[str, BaseFunction] = {}
        self._kwargs = kwargs
        try:
            import xxhash
            self._hash_func = xxhash.xxh64
        except ImportError:
            from hashlib import md5
            self._hash_func = md5
        
        try:
            import cloudpickle
            self._pickle = cloudpickle
        except ImportError:
            import pickle
            self._pickle = pickle

    @property
    def api(self) -> 'OpenAISessionManager':
        """
        Returns the API
        """
        if self._api is None:
            from async_openai.client import OpenAIManager
            self._api = OpenAIManager
        return self._api
    
    @property
    def autologger(self) -> 'Logger':
        """
        Returns the logger
        """
        return self.logger if \
            (self.debug_enabled or self.settings.is_development_env) else self.null_logger

    @property
    def cache(self) -> 'PersistentDict':
        """
        Gets the cache
        """
        if self._cache is None:
            serializer_kwargs = {
                'compression': self._kwargs.get('serialization_compression', None),
                'compression_level': self._kwargs.get('serialization_compression_level', None),
                'raise_errors': True,
            }
            kwargs = {
                'base_key': f'openai.functions.{self.api.settings.app_env.name}.{self.api.settings.proxy.proxy_app_name or "default"}',
                'expiration': self._kwargs.get('cache_expiration', 60 * 60 * 24 * 3),
                'serializer': self._kwargs.get('serialization', 'json'),
                'serializer_kwargs': serializer_kwargs,
            }
            try:
                import kvdb
                self._cache = kvdb.create_persistence(session_name = 'openai', **kwargs)
            except ImportError:
                from lazyops.libs.persistence import PersistentDict
                self._cache = PersistentDict(**kwargs)
        return self._cache

    def register_function(
        self,
        func: Union[BaseFunction, Type[BaseFunction], str],
        name: Optional[str] = None,
        overwrite: Optional[bool] = False,
        raise_error: Optional[bool] = False,
        initialize: Optional[bool] = True,
        **kwargs,
    ):
        """
        Registers the function
        """
        if isinstance(func, str):
            from lazyops.utils.lazy import lazy_import
            func = lazy_import(func)
        if isinstance(func, type) and initialize:
            func = func(**kwargs)
        name = name or func.name
        if not overwrite and name in self.functions:
            if raise_error: raise ValueError(f"Function {name} already exists")
            return
        self.functions[name] = func
        self.autologger.info(f"Registered Function: |g|{name}|e|", colored=True)

    def create_hash(self, *args, **kwargs) -> str:
        """
        Creates a hash
        """
        key = args or ()
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        sorted_items = sorted(kwargs.items())
        for item in sorted_items:
            key += item
        key = ':'.join(str(k) for k in key)
        return self._hash_func(key).hexdigest()
    
    async def acreate_hash(self, *args, **kwargs) -> str:
        """
        Creates a hash
        """
        return await self.api.pooler.asyncish(self.create_hash, *args, **kwargs)
    
    def _get_function(self, name: str) -> Optional[BaseFunction]:
        """
        Gets the function
        """
        func = self.functions.get(name)
        if not func: return None
        if isinstance(func, type):
            func = func(**self._kwargs)
            self.functions[name] = func
        return func

    def get(self, name: Union[str, 'FunctionT']) -> Optional['FunctionT']:
        """
        Gets the function
        """
        return name if isinstance(name, BaseFunction) else self._get_function(name)
        
    
    def parse_iterator_func(
        self,
        function: 'FunctionT',
        *args,
        with_index: Optional[bool] = False,
        **function_kwargs,
    ) -> Tuple[int, Set, Dict[str, Any]]:
        """
        Parses the iterator function kwargs
        """
        func_iter_arg = args[0]
        args = args[1:]
        idx = None
        if with_index: idx, item = func_iter_arg
        else: item = func_iter_arg
        _func_kwargs = function.get_function_kwargs()
        if isinstance(item, dict) and any(k in _func_kwargs for k in item):
            function_kwargs.update(item)
        else:
            # Get the missing function kwargs
            _added = False
            for k in _func_kwargs:
                if k not in function_kwargs:
                    function_kwargs[k] = item
                    # self.autologger.info(f"Added missing function kwarg: {k} = {item}", prefix = function.name, colored = True)
                    _added = True
                    break
            if not _added:
                # If not, then add the item as the first argument
                args = (item,) + args
        return idx, args, function_kwargs

    def execute(
        self,
        function: Union['FunctionT', str],
        *args,
        item_hashkey: Optional[str] = None,
        cachable: Optional[bool] = True,
        overrides: Optional[List[str]] = None,
        with_index: Optional[bool] = False,
        **function_kwargs
    ) -> Union[Optional['FunctionSchemaT'], Tuple[int, Optional['FunctionSchemaT']]]:
        """
        Runs the function
        """
        overwrite = function_kwargs.pop('overwrite', None)
        overwrite = overwrite or overrides and 'functions' in overrides
        function = self.get(function)
        if overwrite and overrides and self.check_value_present(overrides, f'{function.name}.cachable'):
            cachable = False
        
        # Iterators
        is_iterator = function_kwargs.pop('_is_iterator', False)
        if is_iterator:
            idx, args, function_kwargs = self.parse_iterator_func(function, *args, with_index = with_index, **function_kwargs)
        
        if item_hashkey is None: item_hashkey = self.create_hash(**function_kwargs)
        key = f'{item_hashkey}.{function.name}'
        if function.has_diff_model_than_default:
            key += f'.{function.default_model_func}'

        t = Timer()
        result = None
        cache_hit = False
        if self.cache_enabled and not overwrite:
            result: 'FunctionResultT' = self.cache.fetch(key)
            if result:
                if isinstance(result, dict): result = function.schema.model_validate(result)
                result.function_name = function.name
                cache_hit = True
        
        if result is None:
            result = function(*args, cachable = cachable, is_async = False, **function_kwargs)
            if self.cache_enabled and function.is_valid_response(result):
                self.cache.set(key, result)
        
        self.autologger.info(f"Function: {function.name} in {t.total_s} (Cache Hit: {cache_hit})", prefix = key, colored = True)
        if is_iterator and with_index:
            return idx, result if function.is_valid_response(result) else (idx, None)
        return result if function.is_valid_response(result) else None

    
    async def aexecute(
        self,
        function: Union['FunctionT', str],
        *args,
        item_hashkey: Optional[str] = None,
        cachable: Optional[bool] = True,
        overrides: Optional[List[str]] = None,
        with_index: Optional[bool] = False,
        **function_kwargs
    ) -> Union[Optional['FunctionSchemaT'], Tuple[int, Optional['FunctionSchemaT']]]:
        # sourcery skip: low-code-quality
        """
        Runs the function
        """
        overwrite = function_kwargs.pop('overwrite', None)
        overwrite = overwrite or overrides and 'functions' in overrides
        function = self.get(function)
        if overwrite and overrides and self.check_value_present(overrides, f'{function.name}.cachable'):
            cachable = False
        
        # Iterators
        is_iterator = function_kwargs.pop('_is_iterator', False)
        if is_iterator:
            idx, args, function_kwargs = self.parse_iterator_func(function, *args, with_index = with_index, **function_kwargs)

        if item_hashkey is None: item_hashkey = await self.acreate_hash(*args, **function_kwargs)
        key = f'{item_hashkey}.{function.name}'
        if function.has_diff_model_than_default:
            key += f'.{function.default_model_func}'

        t = Timer()
        result = None
        cache_hit = False
        if self.cache_enabled and not overwrite:
            result: 'FunctionResultT' = await self.cache.afetch(key)
            if result:
                if isinstance(result, dict): result = function.schema.model_validate(result)
                result.function_name = function.name
                cache_hit = True
        
        if result is None:
            result = await function(*args, cachable = cachable, is_async = True, **function_kwargs)
            if self.cache_enabled and function.is_valid_response(result):
                await self.cache.aset(key, result)
        
        self.autologger.info(f"Function: {function.name} in {t.total_s} (Cache Hit: {cache_hit})", prefix = key, colored = True)
        if is_iterator and with_index:
            return idx, result if function.is_valid_response(result) else (idx, None)
        return result if function.is_valid_response(result) else None

    
    
    @property
    def function_names(self) -> List[str]:
        """
        Returns the function names
        """
        return list(self.functions.keys())
    
    def __getitem__(self, name: str) -> Optional['FunctionT']:
        """
        Gets the function
        """
        return self.get(name)

    def __setitem__(self, name: str, value: Union[FunctionT, Type[FunctionT], str]):
        """
        Sets the function
        """
        return self.register_function(value, name = name)
    
    def append(self, value: Union[FunctionT, Type[FunctionT], str]):
        """
        Appends the function
        """
        return self.register_function(value)
    

    def check_value_present(
        self, items: List[str], *values: str,
    ) -> bool:
        """
        Checks if the value is present
        """
        if not values:
            return any(self.name in item for item in items)
        for value in values:
            key = f'{self.name}.{value}' if value else self.name
            if any((key in item or value in item) for item in items):
                return True
        return False
    
    def map(
        self,
        function: Union['FunctionT', str],
        iterable_kwargs: List[Union[Dict[str, Any], Any]],
        *args,
        cachable: Optional[bool] = True,
        overrides: Optional[List[str]] = None,
        return_ordered: Optional[bool] = True,
        with_index: Optional[bool] = False,
        **function_kwargs
    ) -> List[Union[Optional['FunctionSchemaT'], Tuple[int, Optional['FunctionSchemaT']]]]:
        """
        Maps the function to the iterable in parallel
        """
        partial = functools.partial(
            self.execute, 
            function, 
            # *args,
            cachable = cachable, 
            overrides = overrides, 
            _is_iterator = True,
            with_index = with_index,
            **function_kwargs
        )
        if with_index: iterable_kwargs = list(enumerate(iterable_kwargs))
        return self.api.pooler.map(partial, iterable_kwargs, *args, return_ordered = return_ordered)
    
    async def amap(
        self,
        function: Union['FunctionT', str],
        iterable_kwargs: List[Dict[str, Any]],
        *args,
        cachable: Optional[bool] = True,
        overrides: Optional[List[str]] = None,
        return_ordered: Optional[bool] = True,
        concurrency_limit: Optional[int] = None,
        with_index: Optional[bool] = False,
        **function_kwargs
    ) -> List[Union[Optional['FunctionSchemaT'], Tuple[int, Optional['FunctionSchemaT']]]]:
        """
        Maps the function to the iterable in parallel
        """
        partial = functools.partial(
            self.aexecute, 
            function, 
            # *args,
            cachable = cachable, 
            overrides = overrides, 
            _is_iterator = True,
            with_index = with_index,
            **function_kwargs
        )
        if with_index: iterable_kwargs = list(enumerate(iterable_kwargs))
        return await self.api.pooler.amap(partial, iterable_kwargs, *args, return_ordered = return_ordered, concurrency_limit = concurrency_limit)
    
    def iterate(
        self,
        function: Union['FunctionT', str],
        iterable_kwargs: List[Dict[str, Any]],
        *args,
        cachable: Optional[bool] = True,
        overrides: Optional[List[str]] = None,
        return_ordered: Optional[bool] = False,
        with_index: Optional[bool] = False,
        **function_kwargs
    ) -> Generator[Union[Optional['FunctionSchemaT'], Tuple[int, Optional['FunctionSchemaT']]], None, None]:
        """
        Maps the function to the iterable in parallel
        """
        partial = functools.partial(
            self.execute, 
            function, 
            # *args, 
            cachable = cachable, 
            overrides = overrides, 
            _is_iterator = True,
            with_index = with_index,
            **function_kwargs
        )
        if with_index: iterable_kwargs = list(enumerate(iterable_kwargs))
        return self.api.pooler.iterate(partial, iterable_kwargs, *args, return_ordered = return_ordered)

    def aiterate(
        self,
        function: Union['FunctionT', str],
        iterable_kwargs: List[Dict[str, Any]],
        *args,
        cachable: Optional[bool] = True,
        overrides: Optional[List[str]] = None,
        return_ordered: Optional[bool] = False,
        concurrency_limit: Optional[int] = None,
        with_index: Optional[bool] = False,
        **function_kwargs
    ) -> AsyncGenerator[Union[Optional['FunctionSchemaT'], Tuple[int, Optional['FunctionSchemaT']]], None]:
        """
        Maps the function to the iterable in parallel
        """
        partial = functools.partial(
            self.aexecute, 
            function, 
            # *args, 
            cachable = cachable, 
            overrides = overrides, 
            _is_iterator = True,
            with_index = with_index,
            **function_kwargs
        )
        if with_index: iterable_kwargs = list(enumerate(iterable_kwargs))
        return self.api.pooler.aiterate(partial, iterable_kwargs, *args, return_ordered = return_ordered, concurrency_limit = concurrency_limit)
    
    def __call__(
        self,
        function: Union['FunctionT', str],
        *args,
        item_hashkey: Optional[str] = None,
        cachable: Optional[bool] = True,
        overrides: Optional[List[str]] = None,
        is_async: Optional[bool] = True,
        **function_kwargs
    ) -> Union[Awaitable['FunctionSchemaT'], 'FunctionSchemaT']:
        """
        Runs the function
        """
        method = self.aexecute if is_async else self.execute
        return method(
            function = function,
            *args,
            item_hashkey = item_hashkey,
            cachable = cachable,
            overrides = overrides,
            **function_kwargs
        )
    




OpenAIFunctions: FunctionManager = ProxyObject(FunctionManager)


import time
import asyncio
from typing import Optional, Type, Any, Union, List, Dict, overload
from lazyops.types import validator, lazyproperty, Field

from async_openai.types.context import ModelContextHandler
from async_openai.types.resources import BaseResource
from async_openai.types.responses import BaseResponse
from async_openai.types.routes import BaseRoute
from async_openai.types.errors import RateLimitError, InvalidMaxTokens, InvalidRequestError, APIError, MaxRetriesExceeded
from async_openai.utils import logger

__all__ = [
    'EmbeddingData',
    'EmbeddingObject',
    'EmbeddingResponse',
    'EmbeddingRoute',
]



class EmbeddingData(BaseResource):
    object: Optional[str] = 'embedding'
    embedding: Optional[List[float]] = []
    index: Optional[int] = 0

class EmbeddingObject(BaseResource):
    model: Optional[str] = "text-embedding-ada-002"
    input: Optional[Union[List[Any], Any]] = None
    user: Optional[str] = None

    @validator('model', pre=True, always=True)
    def validate_model(cls, v, values: Dict[str, Any]) -> str:
        """
        Validate the model
        """
        if not v:
            if values.get('engine'):
                v = values.get('engine')
            elif values.get('deployment'):
                v = values.get('deployment')
        v = ModelContextHandler.resolve_model_name(v)
        # if values.get('validate_model_aliases', False):
        #     v = ModelContextHandler[v].name
        return v
    

    def dict(self, *args, exclude: Any = None, **kwargs):
        """
        Returns the dict representation of the response
        """
        return super().dict(*args, exclude = exclude, **kwargs)
    


class EmbeddingResponse(BaseResponse):
    data: Optional[List[EmbeddingData]] = None
    data_model: Optional[Type[BaseResource]] = EmbeddingData
    input_object: Optional[EmbeddingObject] = None

    @lazyproperty
    def embeddings(self) -> List[List[float]]:
        """
        Returns the text for the response
        object
        """
        if self.data:
            return [data.embedding for data in self.data] if self.data else []
        return None

    @lazyproperty
    def openai_model(self):
        """
        Returns the model for the completions
        """
        return self.headers.get('openai-model', self.input_object.model)

    @lazyproperty
    def consumption(self) -> int:
        """
        Returns the consumption for the completions
        """ 
        return ModelContextHandler.get_consumption_cost(
            model_name = self.openai_model,
            usage = self.usage,
        )



class EmbeddingRoute(BaseRoute):
    input_model: Optional[Type[BaseResource]] = EmbeddingObject
    response_model: Optional[Type[BaseResource]] = EmbeddingResponse
    api_resource: Optional[str] = Field(default = 'embeddings')
    root_name: Optional[str] = Field(default = 'embedding')

    # @lazyproperty
    # def api_resource(self):
    #     return 'embeddings'
    
    # @lazyproperty
    # def root_name(self):
    #     return 'embedding'

    @overload
    def create(
        self, 
        input: Optional[Union[str, List[str], List[List]]] = None,
        model: Optional[str] = "text-embedding-ada-002",
        user: Optional[str] = None,
        auto_retry: Optional[bool] = False,
        auto_retry_limit: Optional[int] = None,
        **kwargs
    ) -> EmbeddingResponse:
        """
        Creates a embedding response for the provided prompt and parameters

        Usage:

        ```python
        >>> result = OpenAI.embedding.create(
        >>>    input = 'say this is a test',
        >>> )
        ```

        **Parameters:**

        :input (string, array, required): Input text to embed, encoded as a string or array of tokens. To embed multiple inputs in a single request, pass an array of strings or array of token arrays. Each input must not exceed the max input tokens for the model (8191 tokens for text-embedding-ada-002). Example Python code for counting tokens.

        :model (string, required): ID of the model to use. You can use the List models API to see all of your available models, or see our Model overview for descriptions of them.
        Default: `text-embedding-ada-002`

        :user (optional): A unique identifier representing your end-user, which can help OpenAI to 
        monitor and detect abuse.
        Default: `None`

        Returns: `EmbeddingResponse`
        """
        ...

    def create(
        self, 
        input_object: Optional[EmbeddingObject] = None,
        auto_retry: Optional[bool] = False,
        auto_retry_limit: Optional[int] = None,
        **kwargs
    ) -> EmbeddingResponse:
        """
        Creates a embedding response for the provided prompt and parameters

        Usage:

        ```python
        >>> result = OpenAI.embedding.create(
        >>>    input = 'say this is a test',
        >>> )
        ```

        **Parameters:**

        :input (string, array, required): Input text to embed, encoded as a string or array of tokens. To embed multiple inputs in a single request, pass an array of strings or array of token arrays. Each input must not exceed the max input tokens for the model (8191 tokens for text-embedding-ada-002). Example Python code for counting tokens.

        :model (string, required): ID of the model to use. You can use the List models API to see all of your available models, or see our Model overview for descriptions of them.
        Default: `text-embedding-ada-002`

        :user (optional): A unique identifier representing your end-user, which can help OpenAI to 
        monitor and detect abuse.
        Default: `None`

        Returns: `EmbeddingResponse`
        """
        if self.is_azure and self.azure_model_mapping and kwargs.get('model') and kwargs['model'] in self.azure_model_mapping:
            kwargs['model'] = self.azure_model_mapping[kwargs['model']]
            kwargs['validate_model_aliases'] = False

        current_attempt = kwargs.pop('_current_attempt', 0)
        if not auto_retry:
            return super().create(input_object=input_object, **kwargs)
        
        # Handle Auto Retry Logic
        if not auto_retry_limit: auto_retry_limit = self.settings.max_retries
        try:
            return super().create(input_object = input_object, **kwargs)
        except RateLimitError as e:
            if current_attempt >= auto_retry_limit:
                raise MaxRetriesExceeded(name = self.name, attempts = current_attempt, base_exception = e) from e
            sleep_interval = e.retry_after_seconds * 1.5 if e.retry_after_seconds else 15.0
            logger.warning(f'[{self.name}: {current_attempt}/{auto_retry_limit}] Rate Limit Error. Sleeping for {sleep_interval} seconds')
            time.sleep(sleep_interval)
            current_attempt += 1
            return self.create(
                input_object = input_object,
                auto_retry = auto_retry,
                auto_retry_limit = auto_retry_limit,
                _current_attempt = current_attempt,
                **kwargs
            )

        
        except APIError as e:
            if current_attempt >= auto_retry_limit:
                raise MaxRetriesExceeded(name = self.name, attempts=current_attempt, base_exception = e) from e
            logger.warning(f'[{self.name}: {current_attempt}/{auto_retry_limit}] API Error: {e}. Sleeping for 10 seconds')
            time.sleep(10.0)
            current_attempt += 1
            return self.create(
                input_object = input_object,
                auto_retry = auto_retry,
                auto_retry_limit = auto_retry_limit,
                _current_attempt = current_attempt,
                **kwargs
            )
        
        except (InvalidMaxTokens, InvalidRequestError) as e:
            raise e
        
        except Exception as e:
            if current_attempt >= auto_retry_limit:
                raise MaxRetriesExceeded(name = self.name, attempts = current_attempt, base_exception = e) from e
            logger.warning(f'[{self.name}: {current_attempt}/{auto_retry_limit}] Unknown Error: {e}. Sleeping for 10 seconds')
            time.sleep(10.0)
            current_attempt += 1
            return self.create(
                input_object = input_object,
                auto_retry = auto_retry,
                auto_retry_limit = auto_retry_limit,
                _current_attempt = current_attempt,
                **kwargs
            )

    @overload
    async def async_create(
        self, 
        input: Optional[Union[str, List[str], List[List]]] = None,
        model: Optional[str] = "text-embedding-ada-002",
        user: Optional[str] = None,
        auto_retry: Optional[bool] = False,
        auto_retry_limit: Optional[int] = None,
        **kwargs
    ) -> EmbeddingResponse:
        """
        Creates a embedding response for the provided prompt and parameters

        Usage:

        ```python
        >>> result = await OpenAI.embedding.async_create(
        >>>    input = 'say this is a test',
        >>> )
        ```

        **Parameters:**

        :input (string, array, required): Input text to embed, encoded as a string or array of tokens. To embed multiple inputs in a single request, pass an array of strings or array of token arrays. Each input must not exceed the max input tokens for the model (8191 tokens for text-embedding-ada-002). Example Python code for counting tokens.

        :model (string, required): ID of the model to use. You can use the List models API to see all of your available models, or see our Model overview for descriptions of them.
        Default: `text-embedding-ada-002`

        :user (optional): A unique identifier representing your end-user, which can help OpenAI to 
        monitor and detect abuse.
        Default: `None`

        Returns: `EmbeddingResponse`
        """
        ...

    async def async_create(
        self, 
        input_object: Optional[EmbeddingObject] = None,
        auto_retry: Optional[bool] = False,
        auto_retry_limit: Optional[int] = None,
        **kwargs
    ) -> EmbeddingResponse:
        """
        Usage:

        ```python
        >>> result = OpenAI.embedding.create(
        >>>    input = 'say this is a test',
        >>> )
        ```

        **Parameters:**

        :input (string, array, required): Input text to embed, encoded as a string or array of tokens. To embed multiple inputs in a single request, pass an array of strings or array of token arrays. Each input must not exceed the max input tokens for the model (8191 tokens for text-embedding-ada-002). Example Python code for counting tokens.

        :model (string): ID of the model to use. You can use the List models API to see all of your available models, or see our Model overview for descriptions of them.
        Default: `text-embedding-ada-002`

        :user (optional): A unique identifier representing your end-user, which can help OpenAI to 
        monitor and detect abuse.
        Default: `None`

        Returns: `EmbeddingResponse`
        """
        if self.is_azure and self.azure_model_mapping and kwargs.get('model') and kwargs['model'] in self.azure_model_mapping:
            kwargs['model'] = self.azure_model_mapping[kwargs['model']]
            kwargs['validate_model_aliases'] = False

        current_attempt = kwargs.pop('_current_attempt', 0)
        if not auto_retry:
            return await super().async_create(input_object = input_object, **kwargs)

        # Handle Auto Retry Logic
        if not auto_retry_limit: auto_retry_limit = self.settings.max_retries
        try:
            return await super().async_create(input_object = input_object, **kwargs)
        except RateLimitError as e:
            if current_attempt >= auto_retry_limit:
                raise MaxRetriesExceeded(name = self.name, attempts = current_attempt, base_exception = e) from e
            sleep_interval = e.retry_after_seconds * 1.5 if e.retry_after_seconds else 15.0
            logger.warning(f'[{self.name}: {current_attempt}/{auto_retry_limit}] Rate Limit Error. Sleeping for {sleep_interval} seconds')
            await asyncio.sleep(sleep_interval)
            current_attempt += 1
            return await self.async_create(
                input_object = input_object,
                auto_retry = auto_retry,
                auto_retry_limit = auto_retry_limit,
                _current_attempt = current_attempt,
                **kwargs
            )
        except APIError as e:
            if current_attempt >= auto_retry_limit:
                raise MaxRetriesExceeded(name = self.name, attempts = current_attempt, base_exception = e) from e
            logger.warning(f'[{self.name}: {current_attempt}/{auto_retry_limit}] API Error: {e}. Sleeping for 10 seconds')
            await asyncio.sleep(10.0)
            current_attempt += 1
            return await self.async_create(
                input_object = input_object,
                auto_retry = auto_retry,
                auto_retry_limit = auto_retry_limit,
                _current_attempt = current_attempt,
                **kwargs
            )

        except (InvalidMaxTokens, InvalidRequestError) as e:
            raise e
        
        except Exception as e:
            if current_attempt >= auto_retry_limit:
                raise MaxRetriesExceeded(name = self.name, attempts = current_attempt, base_exception = e) from e
            logger.warning(f'[{self.name}: {current_attempt}/{auto_retry_limit}] Unknown Error: {e}. Sleeping for 10 seconds')
            await asyncio.sleep(10.0)
            current_attempt += 1
            return await self.async_create(
                input_object = input_object,
                auto_retry = auto_retry,
                auto_retry_limit = auto_retry_limit,
                _current_attempt = current_attempt,
                **kwargs
            )



from typing import Optional, Type, Any, Union, List, Dict
from lazyops.types import validator, lazyproperty, BaseModel, Field

from async_openai.types.options import ImageSize, ImageFormat
from async_openai.types.resources import BaseResource, File, FileType, _has_fileio
from async_openai.types.responses import BaseResponse
from async_openai.types.routes import BaseRoute

__all__ = [
    'ImageData',
    'ImageObject',
    'ImageResponse',
    'ImageRoute',
]

class ImageData(BaseModel):
    url: Optional[str] = None
    data: Optional[bytes] = None

class ImageObject(BaseResource):
    prompt: Optional[str]
    mask: Optional[Union[str, FileType, Any]]
    image: Optional[Union[str, FileType, Any]]
    n: Optional[int] = 1
    size: Optional[Union[str, ImageSize]] = ImageSize.large
    response_format: Optional[Union[str, ImageFormat]] = ImageFormat.url
    user: Optional[str] = None

    @validator("size")
    def validate_size(cls, value):
        return ImageSize.from_str(value) if isinstance(value, str) else value
    
    @validator("response_format")
    def validate_response_format(cls, value):
        if isinstance(value, str):
            value = ImageFormat(value)
        return value
    
    def get_params(self, **kwargs) -> List:
        """
        Transforms the data to the req params
        """
        files = [(k, (None, v)) for k, v in self.dict(exclude_none=True, exclude={'mask', 'image'}).items()]
        if self.mask:
            mask = File(self.mask)
            files.append(("mask", ("mask", mask.read_bytes(), "application/octet-stream")))
        if self.image:
            image = File(self.image)
            files.append(("image", ("image", image.read_bytes(), "application/octet-stream")))
        
        return files
    
    async def async_get_params(self, **kwargs) -> List:
        """
        Transforms the data to the req params
        """
        files = [(k, (None, v)) for k, v in self.dict(exclude_none=True, exclude={'mask', 'image'}).items()]
        if self.mask:
            mask = File(self.mask)
            files.append(("mask", ("mask", (await mask.async_read_bytes() if _has_fileio else mask.read_bytes()), "application/octet-stream")))
        if self.image:
            image = File(self.image)
            files.append(("image", ("image", (await image.async_read_bytes() if _has_fileio else image.read_bytes()), "application/octet-stream")))
        
        return files


class ImageResponse(BaseResponse):
    data: Optional[List[ImageData]]
    data_model: Optional[Type[ImageData]] = ImageData

    @lazyproperty
    def image_urls(self) -> List[str]:
        """
        Returns the list of image urls
        """
        if self.data:
            return [data.url for data in self.data] if self.data else []
        return None
    


class ImageRoute(BaseRoute):
    input_model: Optional[Type[BaseResource]] = ImageObject
    response_model: Optional[Type[BaseResource]] = ImageResponse

    api_resource: Optional[str] = Field(default = 'images')

    # @lazyproperty
    # def api_resource(self):
    #     return 'images'

    def create(
        self, 
        input_object: Optional[Type[BaseResource]] = None,
        **kwargs
    ) -> ImageResponse:
        """
        
        """
        return super().create(input_object = input_object, **kwargs)
    
    async def async_create(
        self, 
        input_object: Optional[Type[BaseResource]] = None,
        **kwargs
    ) -> ImageResponse:
        """
        
        """
        return await super().async_create(input_object = input_object, **kwargs)





import inspect
import aiohttpx
from datetime import datetime, timedelta

from typing import Dict, Optional, Iterator, AsyncIterator, Union
from lazyops.utils.helpers import timed, timer, is_coro_func

__all__ = [
    'is_naive',
    'total_seconds',
    'remove_trailing_slash',
    'full_name',
    'merge_dicts',
    'is_coro_func',
    'timed',
    'timer',
    'parse_stream',
    'aparse_stream',
]


def merge_dicts(x: Dict, y: Dict):
    z = x.copy()
    z.update(y)
    return z


def full_name(func, follow_wrapper_chains=True):
    """
    Return full name of `func` by adding the module and function name.

    If this function is decorated, attempt to unwrap it till the original function to use that
    function name by setting `follow_wrapper_chains` to True.
    """
    if follow_wrapper_chains: func = inspect.unwrap(func)
    return f'{func.__module__}.{func.__qualname__}'

def is_naive(dt: datetime):
    """Determines if a given datetime.datetime is naive."""
    return dt.tzinfo is None or dt.tzinfo.utcoffset(dt) is None


def total_seconds(delta: timedelta):
    """Determines total seconds with python < 2.7 compat."""
    # http://stackoverflow.com/questions/3694835/python-2-6-5-divide-timedelta-with-timedelta
    return (delta.microseconds + (delta.seconds + delta.days * 24 * 3600) * 1e6) / 1e6


def remove_trailing_slash(host: str):
    """
    Removes trailing slash from a host if it exists.
    """
    return host[:-1] if host.endswith("/") else host


def parse_stream_line_bytes(line: bytes) -> Optional[str]:
    """
    Parse a line from a Server-Sent Events stream.
    """
    if line:
        if line.strip() == b"data: [DONE]":
            # return here will cause GeneratorExit exception in urllib3
            # and it will close http connection with TCP Reset
            return None
        if line.startswith(b"data: "):
            line = line[len(b"data: "):]
            return line.decode("utf-8")
        else:
            return None
    return None


def parse_stream_line_string(line: str) -> Optional[str]:
    """
    Parse a line from a Server-Sent Events stream.
    """
    if line:
        if line.strip() == "data: [DONE]":
            # return here will cause GeneratorExit exception in urllib3
            # and it will close http connection with TCP Reset
            return None
        return line[len("data: "):] if line.startswith("data: ") else None
    return None

def parse_stream_line(line: Union[str, bytes]) -> Optional[str]:
    """
    Parse a line from a Server-Sent Events stream.
    """
    if isinstance(line, bytes):
        return parse_stream_line_bytes(line)
    elif isinstance(line, str):
        return parse_stream_line_string(line)
    else:
        raise TypeError("line must be str or bytes")


def parse_stream(response: aiohttpx.Response) -> Iterator[str]:
    """
    Parse a Server-Sent Events stream.
    """
    for line in response.iter_lines():
        _line = parse_stream_line(line)
        if _line is not None:
            yield _line

async def aparse_stream(response: aiohttpx.Response) -> AsyncIterator[str]:
    """
    Parse a Server-Sent Events stream.
    """
    async for line in response.aiter_lines():
        _line = parse_stream_line(line)
        if _line is not None:
            yield _line


import os
import json
import aiohttpx
import datetime
from pydantic import ConfigDict
from pydantic.types import ByteSize
from lazyops.types import BaseModel, validator, lazyproperty
from lazyops.types.models import get_pyd_field_names, pyd_parse_obj, get_pyd_dict, _BaseModel
from lazyops.utils import ObjectDecoder
from async_openai.utils.logs import logger
from async_openai.utils.helpers import aparse_stream, parse_stream

from async_openai.types.options import FilePurpose

from typing import Dict, Optional, Any, List, Type, Union, Tuple, Iterator, AsyncIterator, TYPE_CHECKING


try:
    from fileio import File, FileType
    _has_fileio = True
except ImportError:
    from pathlib import Path as File
    FileType = Union[File, str, os.PathLike]
    _has_fileio = False


__all__ = [
    'BaseResource',
    'Permission',
    'Usage',
    'FileObject',
    'EventObject',
    'FileResource',
]

VALID_SEND_KWARGS = [
    'method',
    'url',
    'content',
    'data',
    'files',
    'json',
    'params',
    'headers',
    'cookies',
    'auth',
    'follow_redirects',
    'timeout',
    'extensions',
]


class Usage(BaseModel):
    prompt_tokens: Optional[int] = 0
    completion_tokens: Optional[int] = 0
    total_tokens: Optional[int] = 0

    @lazyproperty
    def consumption(self) -> int:
        """
        Gets the consumption
        """
        return self.total_tokens
    
    def update(self, usage: Union['Usage', Dict[str, int]]):
        """
        Updates the consumption
        """
        for key in {
            'prompt_tokens',
            'completion_tokens',
            'total_tokens',
        }:
            if not hasattr(self, key):
                setattr(self, key, 0)
            val = usage.get(key, 0) if isinstance(usage, dict) else getattr(usage, key, 0)
            setattr(self, key, getattr(self, key) + val)




class BaseResource(BaseModel):

    """
    Base Object class for resources to
    inherit from
    """

    # model_config = ConfigDict(extra = 'allow', arbitrary_types_allowed = True)
    # def get(self, name, default: Any = None):
    #     """
    #     Get an attribute from the model
    #     """
    #     return getattr(self, name, default)

    if TYPE_CHECKING:
        id: Optional[str]
        file_id: Optional[str]
        fine_tune_id: Optional[str]
        model_id: Optional[str]
        completion_id: Optional[str]
        openai_id: Optional[str]


    @lazyproperty
    def resource_id(self):
        """
        Returns the resource id
        """
        if hasattr(self, 'id'):
            return self.id
        if hasattr(self, 'file_id'):
            return self.file_id
        if hasattr(self, 'fine_tune_id'):
            return self.fine_tune_id
        if hasattr(self, 'model_id'):
            return self.model_id
        if hasattr(self, 'completion_id'):
            return self.completion_id
        return self.openai_id if hasattr(self, 'openai_id') else None
    
    @classmethod
    def parse_obj(
        cls,
        obj: Any,
        strict: Optional[bool] = False,
        from_attributes: Optional[bool] = True,
        **kwargs
    ) -> 'BaseResource':
        """
        Parses an object into the resource
        """
        #return cls(**obj)
        # logger.info(f"Obj: {cls}: {obj}")
        return pyd_parse_obj(cls, obj, strict = strict, from_attributes = from_attributes, **kwargs)
    
    @staticmethod
    def create_resource(
        resource: Type['BaseResource'],
        **kwargs
    ) -> Tuple['BaseResource', Dict]:
        """
        Extracts the resource from the kwargs and returns the resource 
        and the remaining kwargs
        """
        resource_fields = get_pyd_field_names(resource)
        resource_kwargs = {k: v for k, v in kwargs.items() if k in resource_fields}
        return_kwargs = {k: v for k, v in kwargs.items() if k not in resource_fields}
        resource_obj = resource.parse_obj(resource_kwargs)
        return resource_obj, return_kwargs
    
    @classmethod
    def create_many(cls, data: List[Dict]) -> List['BaseResource']:
        """
        Creates many resources
        """
        return [cls.parse_obj(d) for d in data]
    
    @staticmethod
    def handle_json(
        content: Any,
        **kwargs
    ) -> Union[Dict, List]:
        """
        Handles the json response
        """
        return json.loads(content, cls = ObjectDecoder, **kwargs)


    @staticmethod
    def handle_stream(
        response: aiohttpx.Response,
        streaming: Optional[bool] = False,
    ) -> Iterator[Dict]:
        """
        Handles the stream response
        """
        for line in parse_stream(response):
            if not line.strip(): continue
            try:
                yield json.loads(line)
            except Exception as e:
                logger.error(f'Error: {line}: {e}')
    
    @staticmethod
    async def ahandle_stream(
        response: aiohttpx.Response,
        streaming: Optional[bool] = False,
    ) -> AsyncIterator[Dict]:
        """
        Handles the stream response
        """
        async for line in aparse_stream(response):
            if not line.strip(): continue
            try:
                yield json.loads(line)
            except Exception as e:
                logger.error(f'Error: {line}: {e}')


    def __getitem__(self, key: str) -> Any:
        """
        Mimic dict
        """
        return getattr(self, key)


class Permission(BaseResource):
    id: str
    object: str
    created: datetime.datetime
    allow_create_engine: bool
    allow_sampling: bool
    allow_logprobs: bool
    allow_search_indices: bool
    allow_view: bool
    allow_fine_tuning: bool
    organization: str
    group: Optional[str]
    is_blocking: bool

    @property
    def since_seconds(self):
        return (datetime.datetime.now(datetime.timezone.utc) - self.created).total_seconds()


class FileObject(BaseResource):
    id: str
    object: Optional[str] = 'file'
    bytes: Optional[ByteSize]
    created_at: Optional[datetime.datetime]
    filename: Optional[str]
    purpose: Optional[FilePurpose] = FilePurpose.fine_tune

    @validator("created_at")
    def validate_created_at(cls, value):
        return datetime.datetime.fromtimestamp(value, datetime.timezone.utc) if value else value
    
    @classmethod
    def create_many(cls, data: List[Dict]) -> List['FileObject']:
        """
        Creates many resources
        """
        return [cls.parse_obj(d) for d in data]

class EventObject(BaseResource):
    object: Optional[str]
    created_at: Optional[datetime.datetime]
    level: Optional[str]
    message: Optional[str]

    @property
    def since_seconds(self) -> int:
        if self.created_at is None: return -1
        return (datetime.datetime.now(datetime.timezone.utc) - self.created_at).total_seconds()


class FileResource(BaseResource):
    file: Optional[Union[str, FileType, Any]]
    file_id: Optional[str]
    filename: Optional[str] = None
    purpose: FilePurpose = FilePurpose.fine_tune
    model: Optional[str]

    @validator("purpose")
    def validate_purpose(cls, value):
        return FilePurpose.parse_str(value) if isinstance(value, str) else value
    
    def get_params(self, **kwargs) -> List:
        """
        Transforms the data to the req params
        """
        files = [("purpose", (None, self.purpose.value))]
        if self.purpose == FilePurpose.search and self.model:
            files.append(("model", (None, self.model)))
        if self.file:
            file = File(self.file)
            files.append(
                ("file", (self.filename or file.name, file.read_bytes(), "application/octet-stream"))
            )
        return files
    
    async def async_get_params(self, **kwargs) -> List:
        """
        Transforms the data to the req params
        """
        files = [("purpose", (None, self.purpose.value))]
        if self.purpose == FilePurpose.search and self.model:
            files.append(("model", (None, self.model)))
        if self.file:
            file = File(self.file)
            files.append(
                ("file", (self.filename or file.name, (await file.async_read_bytes() if _has_fileio else file.read_bytes()), "application/octet-stream"))
            )
        return files


from __future__ import annotations

"""
OpenAI Cost Functions and Handler
"""
import abc
import tiktoken
from pathlib import Path
from lazyops.types import BaseModel, validator, Field, lazyproperty
from lazyops.libs.proxyobj import ProxyObject
from typing import Optional, Union, Dict, Any, List, Tuple, Type, TYPE_CHECKING

if TYPE_CHECKING:
    from .resources import Usage
    from async_openai.schemas.chat import ChatMessage
    from transformers import PreTrainedTokenizer
    from async_openai.utils.external_config import ExternalProviderSettings

pricing_file_path = Path(__file__).parent.joinpath('pricing.yaml')

class ModelCosts(BaseModel):
    """
    Represents a model's costs
    """
    unit: Optional[int] = 1000
    input: Optional[float] = 0.0
    output: Optional[float] = 0.0
    total: Optional[float] = 0.0


class ModelCostItem(BaseModel):
    """
    Represents a model's Cost Item
    """
    name: str
    aliases: Optional[List[str]] = None
    context_length: Optional[int] = 0
    costs: Optional[ModelCosts] = Field(default_factory=ModelCosts)
    endpoints: Optional[List[str]] = None

    def get_costs(
        self, 
        input_tokens: Optional[int] = None,
        output_tokens: Optional[int] = None,
        total_tokens: Optional[int] = None,
        usage: Optional['Usage'] = None,
        **kwargs
    ) -> float:
        """
        Gets the costs
        """
        if usage is not None:
            input_tokens = usage.prompt_tokens
            output_tokens = usage.completion_tokens
        if kwargs.get('prompt_tokens'):
            input_tokens = kwargs['prompt_tokens']
        if kwargs.get('completion_tokens'):
            output_tokens = kwargs['completion_tokens']
        
        assert input_tokens is not None or output_tokens is not None or total_tokens is not None, "Must provide either input_tokens, output_tokens, or total_tokens"
        if self.costs is None: return 0.0
        cost = 0.0
        if self.costs.input:
            cost += self.costs.input * input_tokens / self.costs.unit
        if self.costs.output:
            cost += self.costs.output * output_tokens / self.costs.unit
        if self.costs.total and total_tokens is not None:
            cost += self.costs.total * total_tokens / self.costs.unit
        return cost
        

class ModelCostHandlerClass(abc.ABC):
    """
    The Model Cost Handler
    """
    def __init__(
        self,
        **kwargs,
    ):
        """
        Initializes the Model Cost Handler
        """
        self._models: Optional[Dict[str, ModelCostItem]] = None
        self._model_aliases: Optional[Dict[str, str]] = None
        self.tokenizers: Optional[Dict[str, tiktoken.Encoding]] = {}

        self.external_models: Optional[Dict[str, ModelCostItem]] = {}
        self.external_model_aliases: Optional[Dict[str, str]] = {}
        self.external_tokenizers: Optional[Dict[str, 'PreTrainedTokenizer']] = {}

    @staticmethod
    def load_models() -> Dict[str, ModelCostItem]:
        """
        Loads the models
        """
        import yaml
        models: Dict[str, Dict[str, Any]] = yaml.safe_load(pricing_file_path.read_text())
        return {k: ModelCostItem(name = k, **v) for k, v in models.items()}

    @property
    def models(self) -> Dict[str, ModelCostItem]:
        """
        Gets the models
        """
        if self._models is None: self._models = self.load_models()
        return self._models
    
    @property
    def model_aliases(self) -> Dict[str, str]:
        """
        Gets the model aliases
        """
        if self._model_aliases is None:
            self._model_aliases = {alias: model for model, item in self.models.items() for alias in item.aliases or []}
        return self._model_aliases


    def get_external_model(self, name: str) -> Optional[ModelCostItem]:
        """
        Gets the model
        """
        if name not in self.external_model_aliases and name not in self.external_models:
            raise KeyError(f"Model {name} not found: {self.external_model_aliases} / {list(self.external_models.keys())}")
            # print(f"Model {name} not found: {self.external_model_aliases} / {self.external_models}")
            # return None
        if name in self.external_model_aliases:
            name = self.external_model_aliases[name]
        return self.external_models[name]
    
    def resolve_external_model_name(self, model_name: str) -> str:
        """
        Resolves the Model Name from the model aliases
        """
        # Try to remove the provider name
        model = self.get_external_model(model_name)
        return model.name

    def resolve_model_name(self, model_name: str) -> str:
        """
        Resolves the Model Name from the model aliases
        """
        # Try to remove the version number
        key = model_name.rsplit('-', 1)[0].strip()
        if key in self.model_aliases:
            self.model_aliases[model_name] = self.model_aliases[key]
        if key in self.models:
            self.model_aliases[model_name] = key
            return key
        raise KeyError(f"Model {model_name} not found")
    
    def __getitem__(self, key: str) -> ModelCostItem:
        """
        Gets a model by name
        """
        if '/' in key: return self.get_external_model(key)
        if key not in self.model_aliases and key not in self.models:
            return self.resolve_model_name(key)
        if key in self.model_aliases:
            key = self.model_aliases[key]
        return self.models[key]
    
    def get(self, key: str, default: Optional[str] = None) -> Optional[ModelCostItem]:
        """
        Gets a model by name
        """
        try:
            return self[key]
        except KeyError:
            if default is None:
                raise KeyError(f"Model {key} not found") from None
            return self[default]
        
    def add_provider(self, provider: 'ExternalProviderSettings'):
        """
        Adds a provider to the handler
        """
        for model in provider.models:
            model_name = f'{provider.name}/{model.name}'
            self.external_models[model_name] = model
            if model.name not in self.external_model_aliases:
                self.external_model_aliases[model.name] = model_name
            for alias in model.aliases or []:
                model_alias = f'{provider.name}/{alias}'
                self.external_model_aliases[model_alias] = model_name
                if alias not in self.external_model_aliases:
                    self.external_model_aliases[alias] = model_name
        
    def add_model(self, model: str, source_model: str):
        """
        Add a model to the handler
        
        Args:
            model (str): The model name
            source_model (str): The source model name
        """
        if model in self.model_aliases or model in self.models:
            return
        
        src_model = self[source_model]
        # Add to the model aliases
        self.model_aliases[model] = src_model.name

    def get_external_tokenizer(self, name: str) -> Optional['PreTrainedTokenizer']:
        """
        Gets the tokenizer
        """
        # Remove the provider name
        if name.count('/') > 1:
            name = name.split('/', 1)[1]
        if name not in self.external_tokenizers:
            try:
                from transformers import AutoTokenizer
            except ImportError as e:
                raise ImportError("transformers is not installed, please install it to use this feature") from e            
            tokenizer = AutoTokenizer.from_pretrained(name)
            self.external_tokenizers[name] = tokenizer
        return self.external_tokenizers[name]

    def get_tokenizer(self, name: str) -> Optional[Union[tiktoken.Encoding, 'PreTrainedTokenizer']]:
        """
        Gets the tokenizer
        """
        # Switch the 35 -> 3.5
        # OpenAI Models don't have / in the name
        if '/' in name: return self.get_external_tokenizer(name)
        if '35' in name: name = name.replace('35', '3.5')    
        if name not in self.tokenizers:
            if name in {'text-embedding-3-small', 'text-embedding-3-large'}:
                enc_name = 'cl100k_base'
                self.tokenizers[name] = tiktoken.get_encoding(enc_name)
            else:
                self.tokenizers[name] = tiktoken.encoding_for_model(name)
        return self.tokenizers[name]
    
    def count_chat_tokens(
        self, 
        messages: List[Union[Dict[str, str], 'ChatMessage']],
        model_name: str,
        reply_padding_token_count: Optional[int] = 3,
        message_padding_token_count: Optional[int] = 4,
        **kwargs
    ) -> int:
        """
        Returns the number of tokens in the chat.
        """
        num_tokens = 0
        tokenizer = self.get_tokenizer(model_name)
        for message in messages:
            if message.get('name'):
                num_tokens -= 1
            num_tokens += message_padding_token_count + len(tokenizer.encode(message.get('content', '')))
        num_tokens += reply_padding_token_count  # every reply is primed with <|start|>assistant<|message|>
        return num_tokens

    def count_tokens(
        self,
        text: Union[str, List[str]],
        model_name: str,
        **kwargs
    ) -> int:
        """
        Returns the number of tokens in the text.
        """
        tokenizer = self.get_tokenizer(model_name)
        return (
            sum(len(tokenizer.encode(t)) for t in text)
            if isinstance(text, list)
            else len(tokenizer.encode(text))
        )
    
    def get_consumption_cost(self, model_name: str, usage: 'Usage', **kwargs) -> float:
        """
        Gets the consumption cost
        """
        # Switch the 35 -> 3.5
        if '35' in model_name: model_name = model_name.replace('35', '3.5')
        model = self[model_name]
        if isinstance(usage, dict):
            from .resources import Usage
            usage = Usage(**usage)
        return model.get_costs(usage = usage, **kwargs)
    
    
    def truncate_to_max_length(self, text: str, model_name: str, context_length: Optional[int] = None, **kwargs) -> str:
        """
        Truncates the text to the max length
        """
        tokenizer = self.get_tokenizer(model_name)
        if context_length is None:
            context_length = self[model_name].context_length
        tokens = tokenizer.encode(text)
        if len(tokens) > context_length:
            tokens = tokens[-context_length:]
            decoded = tokenizer.decode(tokens)
            text = text[-len(decoded):]
        return text




class ModelContextHandlerMetaClass(type):
    """
    The Model Cost Handler
    """

    _models: Optional[Dict[str, ModelCostItem]] = None
    _model_aliases: Optional[Dict[str, str]] = None
    tokenizers: Optional[Dict[str, tiktoken.Encoding]] = {}

    def load_models(cls) -> Dict[str, ModelCostItem]:
        """
        Loads the models
        """
        import yaml
        models: Dict[str, Dict[str, Any]] = yaml.safe_load(pricing_file_path.read_text())
        return {k: ModelCostItem(name = k, **v) for k, v in models.items()}

    @property
    def models(cls) -> Dict[str, ModelCostItem]:
        """
        Gets the models
        """
        if cls._models is None:
            cls._models = cls.load_models()
        return cls._models

    @property
    def model_aliases(cls) -> Dict[str, str]:
        """
        Gets the model aliases
        """
        if cls._model_aliases is None:
            cls._model_aliases = {alias: model for model, item in cls.models.items() for alias in item.aliases or []}
        return cls._model_aliases
    
    def resolve_model_name(cls, model_name: str) -> str:
        """
        Resolves the Model Name from the model aliases
        """
        # Try to remove the version number
        key = model_name.rsplit('-', 1)[0].strip()
        if key in cls.model_aliases:
            cls.model_aliases[model_name] = cls.model_aliases[key]
        if key in cls.models:
            cls.model_aliases[model_name] = key
            return key
        raise KeyError(f"Model {model_name} not found")
    
    def __getitem__(cls, key: str) -> ModelCostItem:
        """
        Gets a model by name
        """
        if key not in cls.model_aliases and key not in cls.models:
            return cls.resolve_model_name(key)
        if key in cls.model_aliases:
            key = cls.model_aliases[key]
        return cls.models[key]
    
    def get(cls, key: str, default: Optional[str] = None) -> Optional[ModelCostItem]:
        """
        Gets a model by name
        """
        try:
            return cls[key]
        except KeyError:
            if default is None:
                raise KeyError(f"Model {key} not found") from None
            return cls[default]
        
    def add_model(cls, model: str, source_model: str):
        """
        Add a model to the handler
        
        Args:
            model (str): The model name
            source_model (str): The source model name
        """
        if model in cls.model_aliases or model in cls.models:
            return
        
        src_model = cls[source_model]
        # Add to the model aliases
        cls.model_aliases[model] = src_model.name


    def get_tokenizer(cls, name: str) -> Optional[tiktoken.Encoding]:
        """
        Gets the tokenizer
        """
        # Switch the 35 -> 3.5
        if '35' in name: name = name.replace('35', '3.5')    
        if name not in cls.tokenizers:
            if name in {'text-embedding-3-small', 'text-embedding-3-large'}:
                enc_name = 'cl100k_base'
                cls.tokenizers[name] = tiktoken.get_encoding(enc_name)
            else:
                cls.tokenizers[name] = tiktoken.encoding_for_model(name)
        return cls.tokenizers[name]
    
    def count_chat_tokens(
        cls, 
        messages: List[Union[Dict[str, str], 'ChatMessage']],
        model_name: str,
        reply_padding_token_count: Optional[int] = 3,
        message_padding_token_count: Optional[int] = 4,
        **kwargs
    ) -> int:
        """
        Returns the number of tokens in the chat.
        """
        num_tokens = 0
        tokenizer = cls.get_tokenizer(model_name)
        for message in messages:
            if message.get('name'):
                num_tokens -= 1
            num_tokens += message_padding_token_count + len(tokenizer.encode(message.get('content', '')))
        num_tokens += reply_padding_token_count  # every reply is primed with <|start|>assistant<|message|>
        return num_tokens

    def count_tokens(
        cls,
        text: Union[str, List[str]],
        model_name: str,
        **kwargs
    ) -> int:
        """
        Returns the number of tokens in the text.
        """
        tokenizer = cls.get_tokenizer(model_name)
        return (
            sum(len(tokenizer.encode(t)) for t in text)
            if isinstance(text, list)
            else len(tokenizer.encode(text))
        )
    
    def get_consumption_cost(cls, model_name: str, usage: 'Usage', **kwargs) -> float:
        """
        Gets the consumption cost
        """
        # Switch the 35 -> 3.5
        if '35' in model_name: model_name = model_name.replace('35', '3.5')
        model = cls[model_name]
        if isinstance(usage, dict):
            from .resources import Usage
            usage = Usage(**usage)
        return model.get_costs(usage = usage, **kwargs)
    
    def resolve_model_name(cls, model_name: str) -> str:
        """
        Resolves the Model Name from the model aliases
        """
        return cls.model_aliases.get(model_name, model_name)
    
    def truncate_to_max_length(cls, text: str, model_name: str, context_length: Optional[int] = None, **kwargs) -> str:
        """
        Truncates the text to the max length
        """
        tokenizer = cls.get_tokenizer(model_name)
        if context_length is None:
            context_length = cls[model_name].context_length

        tokens = tokenizer.encode(text)
        if len(tokens) > context_length:
            tokens = tokens[-context_length:]
            decoded = tokenizer.decode(tokens)
            text = text[-len(decoded):]
        
        return text



class ModelContextHandlerV1(metaclass = ModelContextHandlerMetaClass):
    """
    The Model Cost Handler
    """
    pass


ModelContextHandler: ModelCostHandlerClass = ProxyObject(ModelCostHandlerClass)


from __future__ import annotations

"""
Embedding Utility Helpers

These are borrowed from the `openai` experiments library

- We specifically use lazy loading to avoid runtime errors if the user does not have the required dependencies
"""

from lazyops.libs import lazyload as lz
from lazyops.libs.proxyobj import ProxyObject
from lazyops.types.common import Literal
from typing import Dict, Callable, List, Union, Optional

if lz.TYPE_CHECKING:
    from scipy import spatial
    import numpy as np
    from numpy import ndarray
else:
    spatial = lz.LazyLoad("scipy.spatial")
    np = lz.LazyLoad("numpy")
    
def _initialize_distance_dict(*args, **kwargs) -> Dict[str, Callable[..., float]]:
    """
    Initializes the distance dictionary
    """
    return {
        "cosine": spatial.distance.cosine,
        "L1": spatial.distance.cityblock,
        "L2": spatial.distance.euclidean,
        "Linf": spatial.distance.chebyshev,
    }


distance_metrics: Dict[str, Callable[..., float]] = ProxyObject(obj_getter = _initialize_distance_dict)

MetricT = Literal["cosine", "L1", "L2", "Linf"]


def distances_from_embeddings(
    query_embedding: List[float],
    embeddings: List[List[float]],
    distance_metric: Optional[MetricT] = "cosine",
) -> List[List]:
    """
    Return the distances between a query embedding and a list of embeddings.
    """
    return [
        distance_metrics[distance_metric](query_embedding, embedding)
        for embedding in embeddings
    ]


def indices_of_nearest_neighbors_from_distances(
    distances: 'ndarray',
) -> 'ndarray':
    """
    Return a list of indices of nearest neighbors from a list of distances.
    """
    return np.argsort(distances)